import base64
import copy
import enum
import hashlib
import logging
import typing

import requests

from Crypto.PublicKey import RSA, ECC

import ccc.github
import cfg_mgmt
import cfg_mgmt.util
import ci.util
import model
import model.github

from cfg_mgmt.model import CfgQueueEntry
from model.github import (
    GithubConfig,
    GithubCredentials,
)


ci.log.configure_default_logging()
logger = logging.getLogger(__name__)


class KeyAlgorithm(enum.Enum):
    RSA = enum.auto()
    ECC = enum.auto()


def _determine_key_algorithm(key: str) -> KeyAlgorithm:
    try:
        RSA.import_key(key)
        return KeyAlgorithm.RSA
    except ValueError:
        pass

    try:
        ECC.import_key(key)
        return KeyAlgorithm.ECC
    except ValueError:
        pass

    raise ValueError('Unsupported Key format. Currently, only RSA and ECC keys are supported.')


def _rsa_sha256_fingerprint(key: str) -> str:
    # Returns the fingerprint (as shown in the GitHub UI).
    if not key.startswith('ssh-rsa '):
        return ''
    try:
        # fingerprint is built from the raw decoded public key, so we need to strip the prefix when
        # working with the string
        key = key.lstrip('ssh-rsa ').encode()
        key_raw = base64.b64decode(key)

        key_hash = hashlib.sha256(key_raw).digest()
        fingerprint = base64.b64encode(key_hash).decode()
        # remove '=' used for padding (they are not present in GitHub/ssh-keygen when showing
        # fingerprints)
        fingerprint = fingerprint.rstrip('=')
        return f'SHA256:{fingerprint}'
    except:
        # These fingerprints are intended to make our life easier when looking at the rotation
        # queue, they are not important enough to raise any errors.
        return ''


def _rotate_oauth_token(
    github_api_url: str,
    token_to_rotate: str,
) -> str | None:
    '''Rotate the token for the given credentials iff it is an oAuth token, as (poorly) determined
    by the `gho_` prefix.

    If this function is called for a token other than an oAuth token (e.g.: a personal access token)
    `None` is returned.
    The returned token has the same scopes as the initial token. If you'd like to update the scope
    of the token, you need to re-create it with additional scopes.

    Note: This will only work for oAuth tokens associated with 'git-credential-manager'
    (https://github.com/git-ecosystem/git-credential-manager) due to hardcoded client_id and
    client_secret.
    '''

    if not token_to_rotate.startswith('gho_'):
        logger.warning(
            f"oAuth token '{token_to_rotate}' does not start with 'gho_' - will not update."
        )
        return None
    # client_id and client_"secret" can be hardcoded here as they are the same on all github
    # instances. Also, they are hardcoded in the git credential manager sourcecode on github.com as
    # well, so no actual secrets are leaked here.
    # See https://github.com/git-ecosystem/git-credential-manager/
    # blob/f89105b1ce033e8c2756044d8b195eade2889dac/src/shared/GitHub/GitHubConstants.cs#L14-L17
    client_id = '0120e057bd645470c1ed'
    client_secret = '18867509d956965542b521a529a79bb883344c90'

    request_url = f'{github_api_url}/applications/{client_id}/token'
    session = requests.sessions.Session()
    session.auth = (client_id, client_secret)

    # check token (to catch the case if the token not being associated with the git-credentials-
    # manager)
    resp = session.post(
        request_url,
        json={'access_token': token_to_rotate},
        timeout=(4, 31),
    )
    if not resp.ok:
        # No way for the user to see the oAuth token that belongs to the app, but at least checking
        # whether the given token is currently associated with the app can be done.
        logger.warning(
            f"Given oAuth token '{token_to_rotate[0:len(token_to_rotate)//2]}...' did not pass "
            'GitHub verification. Does the oAuth token belong to the git-credentials-manager '
            'application and is valid?'
        )
        return None

    # fire request that causes the actual refresh (i.e. "rotation") to happen.
    # Note: If successful, old token is immediately invalidated.
    resp = session.patch(
        request_url,
        json={'access_token': token_to_rotate},
        timeout=(4, 31),
    )
    resp.raise_for_status()

    response_content = resp.json()

    return response_content['token']


def rotate_cfg_element(
    cfg_element: model.github.GithubConfig,
    cfg_factory: model.ConfigFactory,
) ->  typing.Tuple[cfg_mgmt.revert_function, dict, model.NamedModelElement]:

    # copy passed cfg_element, since we update in-place.
    raw_cfg = copy.deepcopy(cfg_element.raw)
    cfg_to_rotate = model.github.GithubConfig(
        name=cfg_element.name(), raw_dict=raw_cfg, type_name=cfg_element._type_name
    )

    technical_user_credentials = cfg_to_rotate._technical_user_credentials()

    # retrieve current/"old" public keys before updating to be able to store them in deletion
    # queue
    old_user_data = []
    new_public_keys = dict()
    for i, credential in enumerate(technical_user_credentials):
        if not (private_key := credential.private_key()):
            logger.warn(f'{credential.username()=} did not have private-key - skipping')
            continue

        key_algorithm = _determine_key_algorithm(private_key)

        old_public_key = _corresponding_public_key(
            private_key=credential.private_key(),
            key_algorithm=key_algorithm,
        )

        old_key_fingerprint = _rsa_sha256_fingerprint(old_public_key)
        old_token = ''
        # first, try to update token (if possible, see _rotate_oauth_token())
        try:
            # We rotate by refreshing the secondary token and afterwards switching
            # the primary and secondary tokens. In effect, the old primary token would still be
            # valid until the next rotation happens, but we refresh the token when processing the
            # queue later.
            # This is done to avoid breaking running jobs, since existing tokens are invalidated
            # immediately upon refresh.
            token_to_rotate = credential.secondary_auth_token()
            if token_to_rotate:
                if new_token := _rotate_oauth_token(
                    github_api_url=cfg_to_rotate.api_url(),
                    token_to_rotate=token_to_rotate,
                ):
                    old_token = credential.auth_token()
                    credential.raw['secondary_authToken'] = old_token
                    credential.raw['authToken'] = new_token
                    credential.raw['password'] = new_token
            else:
                logger.info(
                    'No secondary oAuth token provided for credential with username '
                    f"'{credential.username}' of github-config '{cfg_to_rotate.name()}' "
                    '- will not attempt rotation.'
                )
        except requests.exceptions.HTTPError as http_error:
            logger.error(f'Error while trying to refresh oAuth token: {http_error=}')
            if i == 0:
                raise
        except Exception as e:
            logger.error(f'Error while trying to refresh oAuth token: {e}')

            # we only abort here if the first rotation failed (assuming that the token was not
            # refreshed). Otherwise we keep on rotating keys/tokens to prevent the previously
            # succeded rotations from being lost due to the abort.
            if i == 0:
                raise
        config_queue_data = {
            'name': credential.username(),
            'public_key': old_public_key,
            'fingerprint': old_key_fingerprint,
        }
        if old_token:
            config_queue_data['oAuthToken'] = old_token

        old_user_data.append(config_queue_data)

        new_public_keys.update({
            credential.username(): update_user(
                github_config=cfg_to_rotate,
                credential=credential,
                cfg_factory=cfg_factory,
                key_algorithm=key_algorithm,
            )
        })

    secret_id = {'github_users': old_user_data}

    def revert():
        logger.warning(
            f"An error occurred during update of github config '{cfg_to_rotate.name()}', "
            'rolling back'
        )
        # No way to revert oAuth-token rotation. Roll back changes to ssh-keys and return
        for credential in technical_user_credentials:
            if not credential.private_key():
                logger.warn(f'{credential.username()=} did not have private-key - skipping')
                continue

            username = credential.username()
            logger.warning(f"Rolling back changes to user '{username}'")
            new_pub_key = new_public_keys[username]
            gh_api = ccc.github.github_api(
                github_cfg=cfg_to_rotate,
                username=username,
                cfg_factory=cfg_factory,
            )
            for key in gh_api.keys():
                if key.key == new_pub_key:
                    key.delete()
                    logger.info(f"Rollback successful for key '{new_pub_key}'.")
                    break
            else:
                logger.warning(
                    f"New public key for '{username}' not known to github, nothing to revert."
                )

    return revert, secret_id, cfg_to_rotate


def update_user(
    github_config: GithubConfig,
    credential: GithubCredentials,
    cfg_factory: model.ConfigFactory,
    key_algorithm: KeyAlgorithm,
) -> str:
    gh_api = ccc.github.github_api(
        github_cfg=github_config,
        username=credential.username(),
        cfg_factory=cfg_factory,
    )
    private_key, public_key = _create_key_pair(key_algorithm=key_algorithm)
    gh_api.create_key(
        title='Auto-generated key',
        key=public_key,
    )
    credential.raw['privateKey'] = private_key
    return public_key


def delete_config_secret(
    cfg_element: model.github.GithubConfig,
    cfg_factory: model.ConfigFactory,
    cfg_queue_entry: CfgQueueEntry,
) -> model.github.GithubConfig | None:
    raw_cfg = copy.deepcopy(cfg_element.raw)
    cfg_element = model.github.GithubConfig(
        name=cfg_element.name(), raw_dict=raw_cfg, type_name=cfg_element._type_name
    )
    for entry in cfg_queue_entry.secretId['github_users']:
        username = entry['name']
        gh_api = ccc.github.github_api(cfg_element, username=username, cfg_factory=cfg_factory)
        for key in gh_api.keys():
            if key.key == entry['public_key']:
                key.delete()
                break
        else:
            logger.warning(
                f'Old public key for {username} not known to github, nothing to delete.'
            )
        if 'oAuthToken' in entry:
            for credential in cfg_element._technical_user_credentials():
                if credential.secondary_auth_token() == (token := entry['oAuthToken']):
                    credential.raw['secondary_authToken'] = _rotate_oauth_token(
                        github_api_url=cfg_element.api_url(),
                        token_to_rotate=token,
                    )
    return cfg_element


def _corresponding_public_key(
    private_key: str,
    key_algorithm: KeyAlgorithm,
) -> str:
    if key_algorithm is KeyAlgorithm.RSA:
        key = RSA.import_key(private_key)
        public_key_str = key.public_key().export_key(format='OpenSSH').decode('utf-8')
    elif key_algorithm is KeyAlgorithm.ECC:
        key = ECC.import_key(private_key)
        # PEM/OpenSSH keys will be returned as string (not necessarily true for other formats)
        public_key_str = key.public_key().export_key(format='OpenSSH')
    else:
        raise NotImplementedError(key_algorithm)

    return public_key_str


def _create_key_pair(
    key_algorithm: KeyAlgorithm,
) -> typing.Tuple[str, str]:
    '''return a (private-key, public-key)-keypair using the given Algorithm as a pair of strings.

    The private-key is returned in PEM-Format whereas the public key is returned as specified by
    OpenSSH (which is the format expected by GitHub).
    '''
    if key_algorithm is KeyAlgorithm.RSA:
        private_key = RSA.generate(bits=4096)
        private_key_str = private_key.export_key(format='PEM').decode("utf-8")
        public_key_str = private_key.public_key().export_key(format='OpenSSH').decode("utf-8")
    elif key_algorithm is KeyAlgorithm.ECC:
        private_key = ECC.generate(curve='ed25519')
        # PEM/OpenSSH keys will be returned as string (not necessarily true for other formats)
        private_key_str = private_key.export_key(format='PEM')
        public_key_str = private_key.public_key().export_key(format='OpenSSH')
    else:
        raise NotImplementedError(key_algorithm)

    return (
        private_key_str,
        public_key_str
    )



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import model.base

from concourse.model.job import (
    JobVariant,
)
from concourse.model.step import (
    PipelineStep,
    PullRequestNotificationPolicy,
    StepNotificationPolicy,
)
from concourse.model.base import (
    AttributeSpec,
    Trait,
    TraitTransformer,
    ModelBase,
    ScriptType,
)

POLICIES_ATTRIBS = (
    AttributeSpec.optional(
        name='require-label',
        default='reviewed/ok-to-test',
        doc='the label required for PR build to start',
    ),
    AttributeSpec.optional(
        name='replacement-label',
        default='needs/ok-to-test',
        doc='the label set after require-label was removed by PR build',
    ),
)


class PullRequestPolicies(ModelBase):
    @classmethod
    def _attribute_specs(cls):
        return POLICIES_ATTRIBS

    def require_label(self):
        return self.raw.get('require-label')

    def replacement_label(self):
        return self.raw.get('replacement-label')


ATTRIBUTES = (
    AttributeSpec.optional(
        name='policies',
        default={
            'require-label': 'reviewed/ok-to-test',
            'replacement-label': 'needs/ok-to-test',
        },
        doc='configures the policies to apply to pull-requests',
        type=PullRequestPolicies,
    ),
    AttributeSpec.optional(
        name='disable-status-report',
        default=[],
        doc='a list of names of steps which shall not report their status to the pull request.',
        type=list,
    ),
)


class PullRequestTrait(Trait):
    @classmethod
    def _attribute_specs(self):
        return ATTRIBUTES

    def policies(self):
        policies_dict = self.raw['policies']
        return PullRequestPolicies(raw_dict=policies_dict)

    def disable_status_report(self):
        return self.raw.get('disable-status-report')

    def transformer(self):
        return PullRequestTraitTransformer(trait=self)


class PullRequestTraitTransformer(TraitTransformer):
    name = 'pull-request'

    def __init__(self, trait, *args, **kwargs):
        self.trait = trait
        super().__init__(*args, **kwargs)

    def inject_steps(self):
        # declare no dependencies --> run asap, but do not block other steps
        rm_pr_label_step = PipelineStep(
                name='rm_pr_label',
                raw_dict={},
                is_synthetic=True,
                notification_policy=StepNotificationPolicy.NO_NOTIFICATION,
                pull_request_notification_policy=PullRequestNotificationPolicy.NO_NOTIFICATION,
                injecting_trait_name=self.name,
                script_type=ScriptType.PYTHON3
        )
        rm_pr_label_step.set_timeout(duration_string='5m')
        yield rm_pr_label_step

    def process_pipeline_args(self, pipeline_args: JobVariant):
        repo_name = pipeline_args.main_repository().logical_name()

        # convert main-repo to PR
        pr_repo = pipeline_args.pr_repository(repo_name)
        pr_repo._trigger = True

        # patch-in the updated repository
        pipeline_args._repos_dict[repo_name] = pr_repo

        # patch the configured steps so that they do not report their status back to PRs
        for step_name in self.trait.disable_status_report():
            if not pipeline_args.has_step(step_name):
                raise model.base.ModelValidationError(
                    f"Reporting to pull requests was disabled for step '{step_name}', but no step "
                    f"'{step_name}' was found in job '{pipeline_args.variant_name}'"
                )
            step = pipeline_args.step(step_name)
            step._pull_request_notification_policy = PullRequestNotificationPolicy.NO_NOTIFICATION



from reutil import re_filter


def test_re_filter_empty_filter_matches_everything():
    empty_filter = re_filter() # empty filter matches nothing

    assert empty_filter('foo') is True

    empty_filter_with_trans = re_filter(value_transformation=str)

    assert empty_filter_with_trans('foo') is True


def test_re_filter_include_matches():
    include_filter = re_filter(include_regexes=('^aaa.*', '^bbb'))

    assert include_filter('aaa')
    assert include_filter('bbb')

    assert not include_filter('ccc')
    assert not include_filter('bbbb') # require full match


def test_re_filter_exclude():
    exclude_filter = re_filter(exclude_regexes=('^aaa.*', '^bbb'))

    assert not exclude_filter('aaa')
    assert not exclude_filter('bbb')
    assert exclude_filter('fooaaa')
    assert exclude_filter('bbbb') # require full match
    assert exclude_filter('ccc')

    exclude_and_include = re_filter(include_regexes=('^a',), exclude_regexes=('^a',))

    # exclusion has precedence
    assert not exclude_and_include('a')



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import collections.abc
import enum
import random
import re

from urllib.parse import urlparse

import github3.github
import github3.exceptions

from model.base import (
    BasicCredentials,
    NamedModelElement,
    ModelValidationError,
)
import ci.util


class Protocol(enum.Enum):
    SSH = 'ssh'
    HTTPS = 'https'


class GithubConfig(NamedModelElement):
    '''
    Not intended to be instantiated by users of this module
    '''

    def purpose_labels(self):
        return set(self.raw.get('purpose_labels', ()))

    def ssh_url(self):
        if Protocol.SSH not in self.available_protocols():
            raise RuntimeError(f"SSH protocol is not available for Github config '{self.name()}'")
        return self.raw.get('sshUrl')

    def http_url(self):
        if Protocol.HTTPS not in self.available_protocols():
            raise RuntimeError(f"HTTPS protocol is not available for Github config '{self.name()}'")
        return self.raw.get('httpUrl')

    def api_url(self):
        return self.raw.get('apiUrl')

    def matches_api_url(self, api_url):
        return api_url == self.api_url()

    def tls_validation(self):
        return not self.raw.get('disable_tls_validation')

    def preferred_protocol(self):
        return self.available_protocols()[0]

    def available_protocols(self):
        '''Return available git protocols, in order of preference (most preferred protcol first)
        '''
        return [Protocol(value) for value in self.raw.get('available_protocols')]

    def credentials(self, technical_user_name: str = None):
        technical_users = self._technical_user_credentials()
        if technical_user_name:
            for credential in technical_users:
                if credential.username() == technical_user_name:
                    return credential
            raise ModelValidationError(
                f'Did not find technical user "{technical_user_name}" '
                f'for Github config "{self.name()}"'
            )

        return random.choice(technical_users)

    def credentials_with_most_remaining_quota(self):
        credentials = self._technical_user_credentials()
        if len(credentials) < 2:
            return credentials[0]

        if self.hostname() == 'github.com':
            ApiCtor = github3.github.GitHub
            api_kwargs = {}
        else:
            ApiCtor = github3.github.GitHubEnterprise
            api_kwargs = {'url': self.http_url()}

        def rate_limit_remaining(credentials) -> int:
            api = ApiCtor(token=credentials.auth_token(), **api_kwargs)
            try:
                return api.ratelimit_remaining
            except github3.exceptions.ConnectionError:
                return 0

        best_credentials = max(credentials, key=rate_limit_remaining)

        return best_credentials

    def _technical_user_credentials(self) -> collections.abc.Iterable["GithubCredentials"]:
        return [
            GithubCredentials(user) for user in self.raw.get('technical_users')
        ]

    def hostname(self):
        if not (parsed := urlparse(self.http_url())).hostname:
            return None
        return parsed.hostname.lower()

    def matches_hostname(self, host_name):
        return host_name.lower() == self.hostname()

    '''
        repos which the user is for
    '''
    def repo_urls(self) -> list[str]:
        return self.raw.get('repo_urls', ())

    def matches_repo_url(self, repo_url):
        parsed_repo_url = ci.util.urlparse(repo_url)

        if not self.repo_urls():
            return self.matches_hostname(host_name=parsed_repo_url.hostname)

        repo_url = ci.util.urljoin(parsed_repo_url.hostname, parsed_repo_url.path)
        for repo_url_regex in self.repo_urls():
            if re.fullmatch(repo_url_regex, repo_url, re.RegexFlag.IGNORECASE):
                return True

        return False

    def _optional_attributes(self):
        return (
            'httpUrl',
            'purpose_labels',
            'sshUrl',
            'repo_urls',
        )

    def _required_attributes(self):
        return [
            'apiUrl',
            'available_protocols',
            'disable_tls_validation',
            'technical_users',
        ]

    def validate(self):
        super().validate()

        available_protocols = self.available_protocols()
        if len(available_protocols) < 1:
            raise ModelValidationError(
                'At least one available protocol must be configured '
                f"for Github config '{self.name()}'"
            )
        if Protocol.SSH in available_protocols and not self.ssh_url():
            raise ModelValidationError(
                f"SSH url is missing for Github config '{self.name()}'"
            )
        if Protocol.HTTPS in available_protocols and not self.http_url():
            raise ModelValidationError(
                f"HTTP url is missing for Github config '{self.name()}'"
            )

        self.credentials() and self.credentials().validate() # XXX refactor


class GithubCredentials(BasicCredentials):
    '''
    Not intended to be instantiated by users of this module
    '''

    def auth_token(self) -> str:
        tokens = self.raw.get('auth_tokens', None)
        if tokens:
            return random.choice(tokens)
        # fallback to single token
        return self.raw.get('authToken')

    def secondary_auth_token(self) -> str:
        return self.raw.get('secondary_authToken')

    def set_auth_token(self, auth_token):
        self.raw['authToken'] = auth_token

    def private_key(self):
        return self.raw.get('privateKey')

    def email_address(self):
        return self.raw.get('emailAddress')

    def _required_attributes(self):
        required_attribs = set(super()._required_attributes())
        return required_attribs | set(('authToken', 'emailAddress'))

    def _optional_attributes(self):
        return (
            'privateKey',
            'secondary_authToken',
            'recovery-codes',
        )



import dataclasses
import datetime
import enum

import dacite
import dateutil.parser

import cnudie.iter
import dso.cvss
import dso.labels
import ocm
import unixutil.model


@dataclasses.dataclass
class ScanArtifact:
    name: str
    label: dso.labels.SourceScanLabel
    component: ocm.Component
    source: ocm.Source


class Datasource:
    ARTEFACT_ENUMERATOR = 'artefact-enumerator'
    BDBA = 'bdba'
    CHECKMARX = 'checkmarx'
    CLAMAV = 'clamav'
    CC_UTILS = 'cc-utils'
    DIKI = 'diki'


def normalise_artefact_extra_id(
    artefact_extra_id: dict[str, str],
    artefact_version: str=None,
) -> str:
    '''
    generate stable representation of `artefact_extra_id` and remove `version` key if
    the specified version is identical to the given artefact version

    sorted by key in alphabetical order and concatinated following pattern:
    key1:value1_key2:value2_ ...
    '''
    if (version := artefact_extra_id.get('version')) and version == artefact_version:
        artefact_extra_id = artefact_extra_id.copy()
        del artefact_extra_id['version']

    s = sorted(artefact_extra_id.items(), key=lambda items: items[0])
    return '_'.join([':'.join(values) for values in s])


@dataclasses.dataclass(frozen=True)
class LocalArtefactId:
    artefact_name: str | None
    artefact_type: str | None
    artefact_version: str | None = None
    artefact_extra_id: dict = dataclasses.field(default_factory=dict)

    def normalised_artefact_extra_id(
        self,
        remove_duplicate_version: bool=False,
    ) -> str:
        return normalise_artefact_extra_id(
            artefact_extra_id=self.artefact_extra_id,
            artefact_version=self.artefact_version if remove_duplicate_version else None,
        )


class ArtefactKind(enum.StrEnum):
    ARTEFACT = 'artefact'
    RESOURCE = 'resource'
    RUNTIME = 'runtime'
    SOURCE = 'source'


def is_ocm_artefact(artefact_kind: ArtefactKind) -> bool:
    return artefact_kind in (ArtefactKind.RESOURCE, ArtefactKind.SOURCE)


@dataclasses.dataclass(frozen=True)
class ComponentArtefactId:
    component_name: str | None
    component_version: str | None
    artefact: LocalArtefactId | None
    artefact_kind: ArtefactKind = ArtefactKind.ARTEFACT
    references: list['ComponentArtefactId'] = dataclasses.field(default_factory=list)

    def as_frozenset(self) -> frozenset[str]:
        props = (
            self.component_name,
            self.component_version,
            self.artefact_kind,
            frozenset(self.references),
        )

        if self.artefact:
            props += (
                self.artefact.artefact_name,
                self.artefact.artefact_version,
                self.artefact.artefact_type,
                # frozenset(self.artefact.artefact_extra_id.items()),
            )

        return frozenset(props)

    def __hash__(self):
        return hash(self.as_frozenset())

    def __eq__(self, other: 'ComponentArtefactId'):
        return self.as_frozenset() == other.as_frozenset()


def component_artefact_id_from_ocm(
    component: ocm.Component,
    artefact: ocm.Resource | ocm.Source,
) -> ComponentArtefactId:
    local_artefact = LocalArtefactId(
        artefact_name=artefact.name,
        artefact_version=artefact.version,
        artefact_type=artefact.type,
        artefact_extra_id=artefact.extraIdentity,
    )

    if isinstance(artefact, ocm.Resource):
        artefact_kind = ArtefactKind.RESOURCE
    elif isinstance(artefact, ocm.Source):
        artefact_kind = ArtefactKind.SOURCE
    else:
        # should not occur
        raise TypeError(artefact)

    return ComponentArtefactId(
        component_name=component.name,
        component_version=component.version,
        artefact=local_artefact,
        artefact_kind=artefact_kind,
    )


class Datatype:
    STRUCTURE_INFO = 'structure_info'
    LICENSE = 'finding/license'
    VULNERABILITY = 'finding/vulnerability'
    MALWARE_FINDING = 'finding/malware'
    DIKI_FINDING = 'finding/diki'
    CODECHECKS_AGGREGATED = 'codechecks/aggregated'
    OS_IDS = 'os_ids'
    RESCORING = 'rescorings'
    COMPLIANCE_SNAPSHOTS = 'compliance/snapshots'
    ARTEFACT_SCAN_INFO = 'meta/artefact_scan_info'


@dataclasses.dataclass(frozen=True)
class Metadata:
    datasource: str
    type: str
    creation_date: datetime.datetime | str = None
    last_update: datetime.datetime | str | None = None


@dataclasses.dataclass(frozen=True)
class OsID:
    os_info: unixutil.model.OperatingSystemId


@dataclasses.dataclass(frozen=True)
class CodecheckFindings:
    high: int
    medium: int
    low: int
    info: int


@dataclasses.dataclass(frozen=True)
class CodecheckSummary:
    findings: CodecheckFindings
    risk_rating: int
    risk_severity: int
    overview_url: str
    report_url: str | None


@dataclasses.dataclass(frozen=True)
class BDBAMixin:
    package_name: str
    package_version: str | None # bdba might be unable to determine a version
    base_url: str
    report_url: str
    product_id: int
    group_id: int


@dataclasses.dataclass(frozen=True)
class License:
    name: str


@dataclasses.dataclass(frozen=True)
class FilesystemPathEntry:
    path: str
    type: str


@dataclasses.dataclass(frozen=True)
class FilesystemPath:
    path: list[FilesystemPathEntry]
    digest: str


@dataclasses.dataclass(frozen=True)
class StructureInfo(BDBAMixin):
    licenses: list[License]
    filesystem_paths: list[FilesystemPath]

    @property
    def key(self) -> str:
        return f'{self.package_name}|{self.package_version}'


@dataclasses.dataclass(frozen=True)
class Finding:
    '''
    Base class for artefact metadata which is interpreted as a finding. "Finding" as in it has a
    severity and might become object of being rescored.
    '''
    severity: str


@dataclasses.dataclass(frozen=True)
class LicenseFinding(Finding, BDBAMixin):
    license: License

    @property
    def key(self) -> str:
        return f'{self.package_name}|{self.package_version}|{self.license.name}'


@dataclasses.dataclass(frozen=True)
class VulnerabilityFinding(Finding, BDBAMixin):
    cve: str
    cvss_v3_score: float
    cvss: dso.cvss.CVSSV3 | dict
    summary: str | None

    @property
    def key(self) -> str:
        return f'{self.package_name}|{self.package_version}|{self.cve}'


@dataclasses.dataclass(frozen=True)
class RescoringVulnerabilityFinding:
    package_name: str
    cve: str

    @property
    def key(self) -> str:
        return f'{self.package_name}|{self.cve}'


@dataclasses.dataclass(frozen=True)
class RescoringLicenseFinding:
    package_name: str
    license: License

    @property
    def key(self) -> str:
        return f'{self.package_name}|{self.license.name}'


@dataclasses.dataclass(frozen=True)
class MalwareFindingDetails:
    filename: str
    content_digest: str
    malware: str
    context: str | None # optional context information, e.g. layer-digest or bucket-id

    @property
    def key(self) -> str:
        return f'{self.content_digest}|{self.filename}|{self.malware}'


@dataclasses.dataclass(frozen=True)
class ClamAVMalwareFinding(Finding):
    finding: MalwareFindingDetails
    octets_count: int
    scan_duration_seconds: float
    clamav_version: str | None
    signature_version: int | None
    freshclam_timestamp: datetime.datetime | None

    @property
    def key(self) -> str:
        return self.finding.key


@dataclasses.dataclass(frozen=True)
class DikiCheck:
    message: str
    targets: list[dict] | dict[str, list[dict]]


@dataclasses.dataclass(frozen=True)
class DikiFinding(Finding):
    provider_id: str
    ruleset_id: str
    ruleset_version: str
    rule_id: str
    checks: list[DikiCheck]

    @property
    def key(self) -> str:
        return f'{self.provider_id}|{self.ruleset_id}:{self.ruleset_version}|{self.rule_id}'


@dataclasses.dataclass(frozen=True)
class User:
    username: str
    type: str = 'user'

    @property
    def key(self) -> str:
        return f'{self.username}|{self.type}'


@dataclasses.dataclass(frozen=True, kw_only=True)
class BDBAUser(User):
    email: str
    firstname: str
    lastname: str
    type: str = 'bdba-user'


@dataclasses.dataclass(frozen=True, kw_only=True)
class GitHubUser(User):
    github_hostname: str
    type: str = 'github-user'


class MetaRescoringRules(enum.StrEnum):
    BDBA_TRIAGE = 'bdba-triage'
    CUSTOM_RESCORING = 'custom-rescoring'
    ORIGINAL_SEVERITY = 'original-severity'


@dataclasses.dataclass(frozen=True)
class CustomRescoring:
    finding: (
        RescoringVulnerabilityFinding
        | RescoringLicenseFinding
        | MalwareFindingDetails
    )
    referenced_type: str
    severity: str
    user: (
        BDBAUser
        | GitHubUser
        | User
    )
    matching_rules: list[str] = dataclasses.field(default_factory=list)
    comment: str | None = None

    @property
    def key(self) -> str:
        return (
            f'{self.referenced_type}|{self.severity}|{self.user.key}|'
            f'{self.comment}|{self.finding.key}'
        )


class ComplianceSnapshotStatuses(enum.StrEnum):
    ACTIVE = 'active'
    INACTIVE = 'inactive'


@dataclasses.dataclass(frozen=True)
class ComplianceSnapshotState:
    timestamp: datetime.datetime
    status: ComplianceSnapshotStatuses | str | int | None = None
    service: str | None = None


@dataclasses.dataclass(frozen=True)
class ComplianceSnapshot:
    cfg_name: str
    latest_processing_date: datetime.date
    correlation_id: str
    state: list[ComplianceSnapshotState]

    @property
    def key(self) -> str:
        return f'{self.cfg_name}|{self.correlation_id}'

    def current_state(
        self,
        service: str = None,
    ) -> ComplianceSnapshotState | None:
        for state in sorted(self.state, key=lambda s: s.timestamp, reverse=True):
            if service == state.service:
                return state
        return None

    def purge_old_states(
        self,
        service: str = None,
    ):
        current_state = None
        for state in sorted(self.state, key=lambda s: s.timestamp, reverse=True):
            if not service == state.service:
                continue

            if not current_state:
                current_state = state
                continue

            self.state.remove(state)


@dataclasses.dataclass(frozen=True)
class ArtefactMetadata:
    '''
    Model class to interact with entries of the delivery-db. In the first place, these entries are
    being identified via `ComponentArtefactId` (`artefact` property) as well as their `Datatype`
    (`meta.type` property). If there might be multiple entries for this tuple, the `data` object
    must define an extra `key` property, which allows a unique identification together with the
    tuple of `artefact` and `meta.type`.

    If an instance of a datatype should become object of being rescored, the `data` property must
    derive from the `Finding` class and implement the `severity` property. Also, a corresponding
    rescoring finding type must be implemented. Apart from the `key` and `severity` property, the
    `data` object may have an arbitrary structure.
    '''
    artefact: ComponentArtefactId
    meta: Metadata
    data: (
        StructureInfo
        | LicenseFinding
        | VulnerabilityFinding
        | ClamAVMalwareFinding
        | DikiFinding
        | CodecheckSummary
        | OsID
        | CustomRescoring
        | ComplianceSnapshot
        | dict # fallback, there should be a type
    )
    id: int | None = None
    discovery_date: datetime.date | None = None # required for finding specific SLA tracking

    @staticmethod
    def from_dict(raw: dict):
        return dacite.from_dict(
            data_class=ArtefactMetadata,
            data=raw,
            config=dacite.Config(
                type_hooks={
                    datetime.datetime: dateutil.parser.isoparse,
                    datetime.date: lambda date: datetime.datetime.fromisoformat(date).date(),
                },
                cast=[
                    ArtefactKind,
                    ComplianceSnapshotStatuses,
                    MetaRescoringRules,
                ],
            ),
        )


def artefact_scan_info(
    artefact_node: cnudie.iter.ArtefactNode,
    datasource: str,
    data: dict={},
) -> ArtefactMetadata:
    '''
    The `data` property may contain extra information about the scan, e.g. a reference to the scan.

    Predefined `data` property for BDBA scan infos:

    data:
        report_url <str>
    '''
    now = datetime.datetime.now()

    artefact_ref = component_artefact_id_from_ocm(
        component=artefact_node.component,
        artefact=artefact_node.artefact,
    )

    meta = Metadata(
        datasource=datasource,
        type=Datatype.ARTEFACT_SCAN_INFO,
        creation_date=now,
        last_update=now,
    )

    return ArtefactMetadata(
        artefact=artefact_ref,
        meta=meta,
        data=data,
    )



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0

import typing

import ci.util
from model.base import (
    NamedModelElement,
    TokenCredentials,
)


class BDBAConfig(NamedModelElement):
    '''
    Not intended to be instantiated by users of this module
    '''
    def credentials(self) -> TokenCredentials:
        return TokenCredentials(
            raw_dict={
                'token': self.raw['credentials']['token'],
            }
        )

    def api_url(self) -> str:
        return self.raw.get('api_url')

    def tls_verify(self) -> bool:
        return self.raw.get('tls_verify', True)

    def group_ids(self) -> list[int]:
        return self.raw.get('group_ids', [])

    def base_url(self) -> str:
        if not (parsed := ci.util.urlparse(self.api_url())):
            return None
        return f'{parsed.scheme}://{parsed.hostname}'

    def matches(
        self,
        group_id: int=None,
        base_url: str=None,
    ) -> int:
        '''
        returns integer in range [-1;2] representing quality of match where -1 states "no match" and
        2 "perfect match".
        '''
        def normalise_url(url: str) -> str:
            parsed = ci.util.urlparse(url)
            return f'{parsed.scheme}://{parsed.hostname}{parsed.path}'

        score = 0

        if base_url:
            base_url = normalise_url(base_url)
            if base_url.startswith(self.base_url()):
                score += 1
            else:
                return -1

        if group_id:
            if group_id in self.group_ids():
                score += 1
            else:
                if self.group_ids():
                    return -1

        return score

    def _required_attributes(self):
        yield from super()._required_attributes()
        yield from (
            'api_url',
            'credentials',
        )

    def validate(self):
        # call credentials method to validate credentials
        self.credentials()


def find_config(
    group_id: int | None,
    base_url: str | None,
    config_candidates: typing.Iterable[BDBAConfig],
) -> BDBAConfig | None:
    matching_cfgs = [
        cfg
        for cfg in config_candidates
        if not cfg.matches(
            group_id=group_id,
            base_url=base_url,
        ) == -1
    ]

    if not matching_cfgs:
        return None

    matching_cfgs = sorted(matching_cfgs, key=lambda c:c.matches(
        group_id=group_id,
        base_url=base_url,
    ))

    # sorted, last element of greatest match quality
    return matching_cfgs[-1]



import datetime
import dataclasses
import urllib.parse
import typing as t


from dacite import from_dict
import requests

import ci.util
import checkmarx.model as cxmodel
import model.checkmarx


def require_auth(f: callable):
    def wrapper(checkmarx_client: 'CheckmarxClient', *args, **kwargs):
        checkmarx_client._auth()
        res = f(checkmarx_client, *args, **kwargs)
        return res

    return wrapper


class CXNotOkayException(Exception):
    def __init__(self, res: requests.Response, msg: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.res = res
        self.msg = msg

    def __repr__(self):
        return f'CXNotOkayException: {self.msg}'


class CheckmarxRoutes:
    '''
    Checkmarx REST API endpoints for the checkmarx base URL.
    '''

    def __init__(self, base_url: str):
        self.base_url = base_url

    def _api_url(self, *parts):
        return ci.util.urljoin(self.base_url, 'cxrestapi', *parts)

    def _web_url(self, *parts):
        return ci.util.urljoin(self.base_url, 'cxwebclient', *parts)

    def auth(self):
        return self._api_url('auth', 'identity', 'connect', 'token')

    def projects(self):
        return self._api_url('projects')

    def project_by_id(self, project_id: int):
        return ci.util.urljoin(self.projects(), str(project_id))

    def scan(self):
        return self._api_url('sast', 'scans')

    def scan_by_id(self, scan_id: int):
        return ci.util.urljoin(self.scan(), str(scan_id))

    def scan_statistics(self, scan_id: int):
        return ci.util.urljoin(self.scan(), str(scan_id), 'resultsStatistics')

    def upload_zipped_source(self, project_id: int):
        return ci.util.urljoin(str(self.project_by_id(project_id)), 'sourceCode', 'attachments')

    def remote_settings_git(self, project_id: int):
        return ci.util.urljoin(self.scan_by_id(project_id), 'sourceCode', 'remoteSettings', 'git')

    def web_ui_scan_history(self, project_id: int):
        query = urllib.parse.urlencode({'id': project_id, 'ProjectState': 'true'})
        return ci.util.urljoin(self._web_url(), 'projectscans.aspx?' + query)

    def web_ui_scan_viewer(self, scan_id: int, project_id: int):
        query = urllib.parse.urlencode({'scanId': scan_id, 'ProjectID': project_id})
        return ci.util.urljoin(self._web_url(), 'ViewerMain.aspx?' + query)


class CheckmarxClient:
    def __init__(self, checkmarx_cfg: model.checkmarx.CheckmarxConfig):
        self.routes = CheckmarxRoutes(base_url=checkmarx_cfg.base_url())
        self.config = checkmarx_cfg
        self.auth = None
        self.session = requests.sessions.Session()

    def _auth(self):
        if self.auth and self.auth.is_valid():
            return self.auth

        creds = self.config.credentials()
        res = self.session.post(
            self.routes.auth(),
            data={
                'username': creds.qualified_username(),
                'password': creds.passwd(),
                'client_id': creds.client_id(),
                'client_secret': creds.client_secret(),
                'scope': creds.scope(),
                'grant_type': 'password',
            },
            timeout=(4, 31),
        )
        res.raise_for_status()

        res = cxmodel.AuthResponse(**res.json())
        res.expires_at = datetime.datetime.fromtimestamp(
            datetime.datetime.now().timestamp() + res.expires_in - 10
        )
        self.auth = res
        return res

    @require_auth
    def request(
        self,
        method: str,
        api_version: str = '1.0',
        print_error: bool = True,
        *args,
        **kwargs,
    ):
        headers = kwargs.pop('headers', {})
        headers['Authorization'] = f'Bearer {self.auth.access_token}'
        if 'Accept' not in headers:
            headers['Accept'] = f'application/json;v={api_version}'

        try:
            timeout = kwargs.pop('timeout')
        except KeyError:
            timeout = (4, 121)

        res = self.session.request(
            method=method,
            headers=headers,
            timeout=timeout,
            *args,
            **kwargs,
        )

        if not res.ok:
            msg = f'{method} request to {res.url=} failed with {res.status_code=} {res.reason=}'
            if print_error:
                ci.util.error(f'Request failed {res.status_code=} {res.text=} {msg=}')

            raise CXNotOkayException(res=res, msg=msg)
        return res

    def create_project(self, name: str, owning_team: str, is_public: bool):
        res = self.request(
            method='POST',
            url=self.routes.projects(),
            json={
                "name": name,
                "owningTeam": owning_team,
                "isPublic": is_public,
            },
        )
        return res

    def upload_zipped_source_code(self, project_id: int, zipped_source):
        res = self.request(
            method='POST',
            url=self.routes.upload_zipped_source(project_id),
            headers={
                'Accept': 'application/json',
            },
            files={'zippedSource': zipped_source},
        )
        return res

    def get_projects(self, team_id: int) -> t.List[cxmodel.ProjectDetails]:
        res = self.request(
            method='GET',
            url=self.routes.projects(),
            params={
                'teamId': team_id,
            },
            print_error=False,
        )
        projects = res.json()
        return [from_dict(data_class=cxmodel.ProjectDetails, data=p) for p in projects]

    def get_project_id_by_name(self, project_name: str, team_id: int):
        res = self.request(
            method='GET',
            url=self.routes.projects(),
            params={
                'projectName': project_name,
                'teamId': team_id,
            },
            print_error=False,
        )
        return res.json()[0].get('id')

    def get_project_by_id(self, project_id: int):
        res = self.request(
            method='GET',
            url=self.routes.project_by_id(project_id=project_id),
            api_version="application/json;v=2.0",
        )
        return from_dict(data_class=cxmodel.ProjectDetails, data=res.json())

    def update_project(self, project_details: cxmodel.ProjectDetails):
        res = self.request(
            method="PUT",
            url=self.routes.project_by_id(project_details.id),
            json={
                'name': project_details.name,
                'owningTeam': project_details.teamId,
                'customFields': [dataclasses.asdict(cf) for cf in project_details.customFields],
            },
        )
        return res

    def start_scan(self, scan_settings: cxmodel.ScanSettings):
        res = self.request(
            method='POST',
            url=self.routes.scan(),
            json=dataclasses.asdict(scan_settings),
        )
        scan_id = res.json()['id']
        return scan_id

    def get_last_scans_of_project(self, project_id: int, last_scans: int = 1):
        res = self.request(
            method="GET",
            url=self.routes.scan(),
            params={
                "last": last_scans,
                "projectId": str(project_id),
            },
            api_version='application/json;v=1.0',
        )
        return [
            from_dict(data_class=cxmodel.ScanResponse, data=resp)
            for resp in res.json()
        ]

    def get_scan_state(self, scan_id: int):
        res = self.request(
            method='GET',
            url=self.routes.scan_by_id(scan_id=scan_id),
        )
        return from_dict(data_class=cxmodel.ScanResponse, data=res.json())

    def get_scan_statistics(self, scan_id: int):
        res = self.request(
            method='GET',
            url=self.routes.scan_statistics(scan_id=scan_id)
        )
        return from_dict(data_class=cxmodel.ScanStatistic, data=res.json())



# SPDX-FileCopyrightText: 2021 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0

import pytest

import ctt.processors as processors
import ctt.processing_model as processing_model


@pytest.fixture
def job():
    return processing_model.ProcessingJob(
        component=None,
        resource=None,
        upload_request=processing_model.ContainerImageUploadRequest(
            source_ref='source:ref',
            target_ref='target:ref',
            remove_files=(),
        ),
    )


def test_noop_processor(job):
    examinee = processors.NoOpProcessor()

    result = examinee.process(job)

    assert result is job


def test_filefilter_processor(job, tmpdir):
    filter_file = tmpdir.join('filters')
    filter_file.write('remove/me')

    examinee = processors.FileFilter(
        filter_files=[filter_file],
    )

    result = examinee.process(job)
    upload_request = result.upload_request
    remove_files = upload_request.remove_files

    assert remove_files == ('remove/me',)



import logging

import ci.log
import ci.util
import ctx
import delivery.client

ci.log.configure_default_logging()
logger = logging.getLogger(__name__)


def _current_cfg_set(
    cfg_factory=None,
):
    if not cfg_factory:
        cfg_factory = ctx.cfg_factory()

    cfg_set_name = ci.util.current_config_set_name()
    cfg_set = cfg_factory.cfg_set(cfg_set_name)

    return cfg_set


def default_client_if_available(
    cfg_factory=None,
) -> delivery.client.DeliveryServiceClient:
    if not cfg_factory:
        cfg_factory = ctx.cfg_factory()

    delivery_endpoints = None

    if ci.util._running_on_ci():
        cfg_set = _current_cfg_set(
            cfg_factory=cfg_factory,
        )

        try:
            delivery_endpoints = cfg_set.delivery_endpoints()
        except ValueError:
            return None
    else:
        delivery_cfg_name = ctx.cfg.ctx.delivery_cfg_name
        if delivery_cfg_name:
            delivery_endpoints = cfg_factory.delivery_endpoints(delivery_cfg_name)

    if not delivery_endpoints:
        return None

    routes = delivery.client.DeliveryServiceRoutes(
        base_url=delivery_endpoints.base_url(),
    )
    return delivery.client.DeliveryServiceClient(routes=routes)


def client(
    cfg_name: str=None,
    cfg_factory=None,
) -> delivery.client.DeliveryServiceClient:
    if not cfg_factory:
        cfg_factory = ctx.cfg_factory()

    if not cfg_name:
        if delivery_client := default_client_if_available():
            return delivery_client
        raise ValueError('no (default) delivery-client could be determined - pass cfg_name')

    delivery_endpoints = cfg_factory.delivery_endpoints(cfg_name)
    routes = delivery.client.DeliveryServiceRoutes(
        base_url=delivery_endpoints.base_url(),
    )

    return delivery.client.DeliveryServiceClient(routes=routes)


def endpoints(cfg_set=None):
    if not ci.util._running_on_ci() and not cfg_set:
        return None

    if not cfg_set:
        cfg_set = _current_cfg_set()

    delivery_endpoints = cfg_set.delivery_endpoints()

    return delivery_endpoints



import dataclasses
import enum
import os
import random
import time

from enum import Enum, IntEnum
from concurrent.futures import ThreadPoolExecutor
import functools
import logging
import textwrap
import threading
import traceback
import typing
import requests.exceptions

import mako.exceptions
import mako.template

from ci.util import (
    existing_dir,
    merge_dicts,
    urljoin,
)
from mailutil import _send_mail
import github.codeowners

import concourse.client.api
from concourse.factory import DefinitionFactory, RawPipelineDefinitionDescriptor
from concourse.enumerator import (
    DefinitionDescriptor,
    DefinitionDescriptorPreprocessor,
    TemplateRetriever,
    GithubOrganisationDefinitionEnumerator,
)

import ccc.concourse
import ccc.github
import concourse.client.model
import concourse.paths
import makoutil
import model.concourse

logger = logging.getLogger(__name__)


class RenderOrigin(enum.Enum):
    LOCAL = 'local'
    WEBHOOK_DISPATCHER = 'webhook dispatcher'
    PIPELINE_REPLICATION = 'pipeline replication'
    UNKNOWN = 'unknown'


def replicate_pipelines(
    cfg_set,
    job_mapping,
    template_path=concourse.paths.template_dir,
    template_include_dir=concourse.paths.template_include_dir,
    unpause_pipelines: bool=True,
    expose_pipelines: bool=True,
    unpause_new_pipelines: bool=True,
    remove_pipelines_filter: typing.Callable[[str], bool]=None,
):
    '''
    @param remove_pipelines_filter: pipeline-names the filter does not match are never removed
    '''

    definition_enumerators = [
        GithubOrganisationDefinitionEnumerator(
            job_mapping=job_mapping,
            cfg_set=cfg_set,
            repository_filter=lambda repo: not repo.archived, # exclude archived repositories
        ),
    ]

    preprocessor = DefinitionDescriptorPreprocessor()
    template_retriever = TemplateRetriever(template_path=template_path)
    renderer = Renderer(
        template_retriever=template_retriever,
        template_include_dir=template_include_dir,
        cfg_set=cfg_set,
        render_origin=RenderOrigin.PIPELINE_REPLICATION,
    )

    deployer = ConcourseDeployer(
        cfg_set=cfg_set,
        unpause_pipelines=unpause_pipelines,
        unpause_new_pipelines=unpause_new_pipelines,
        expose_pipelines=expose_pipelines,
    )

    result_processor = ReplicationResultProcessor(
        cfg_set=cfg_set,
        unpause_new_pipelines=unpause_new_pipelines,
        job_mapping=job_mapping,
        remove_pipelines_filter=remove_pipelines_filter,
    )

    replicator = PipelineReplicator(
        definition_enumerators=definition_enumerators,
        descriptor_preprocessor=preprocessor,
        definition_renderer=renderer,
        definition_deployer=deployer,
        result_processor=result_processor,
    )

    return replicator.replicate()


@functools.lru_cache
def _cc_utils_version():
    try:
        with open(concourse.paths.last_released_tag_file) as f:
            return f.read().strip()
    except:
        return None


class Renderer:
    def __init__(
        self,
        cfg_set,
        template_retriever: TemplateRetriever=TemplateRetriever(),
        template_include_dir=None,
        render_origin: RenderOrigin = RenderOrigin.UNKNOWN,
    ):
        self.template_retriever = template_retriever
        self.render_origin = render_origin
        if template_include_dir:
            template_include_dir = os.path.abspath(template_include_dir)
            self.template_include_dir = os.path.abspath(template_include_dir)
            from mako.lookup import TemplateLookup
            self.lookup = TemplateLookup([template_include_dir])
            self.cfg_set = cfg_set

    def render(
        self,
        definition_descriptor: DefinitionDescriptor,
    ):
        try:
            definition_descriptor = self._render(definition_descriptor)
            logger.info('rendered pipeline {pn}'.format(pn=definition_descriptor.pipeline_name))
            return RenderResult(
                definition_descriptor,
                render_status=RenderStatus.SUCCEEDED,
            )
        except Exception as e:
            logger.warning(
                f"erroneous pipeline definition '{definition_descriptor.pipeline_name}' "
                f"in repository '{definition_descriptor.main_repo.get('path')}' on branch "
                f"'{definition_descriptor.main_repo.get('branch')}'"
            )
            traceback.print_exc()
            return RenderResult(
                definition_descriptor,
                render_status=RenderStatus.FAILED,
                error_details=traceback.format_exc(),
                exception=e,
            )

    def _render(self, definition_descriptor):
        effective_definition = definition_descriptor.pipeline_definition

        # handle inheritance
        for override in definition_descriptor.override_definitions:
            effective_definition = merge_dicts(effective_definition, override)

        template_name = definition_descriptor.template_name()
        template_contents = self.template_retriever.template_contents(template_name)

        pipeline_name = definition_descriptor.pipeline_name

        pipeline_definition = RawPipelineDefinitionDescriptor(
            name=pipeline_name,
            base_definition=effective_definition.get('base_definition', {}),
            jobs=effective_definition.get('jobs', {}),
            template=template_name,
        )

        factory = DefinitionFactory(
            raw_definition_descriptor=pipeline_definition,
            cfg_set=self.cfg_set,
        )
        pipeline_metadata = {
            'definition': factory.create_pipeline_definition(),
            'name': pipeline_definition.name,
            'target_team': definition_descriptor.concourse_target_team,
            'secret_cfg': definition_descriptor.secret_cfg,
            'job_mapping': definition_descriptor.job_mapping,
            'render_origin': self.render_origin.value,
            'cc_utils_version': _cc_utils_version(),
            'pipeline_definition_committish': definition_descriptor.pipeline_definition_committish,
        }

        # also pass pipeline name if this was rendered by a replication job. Will be printed
        # in the meta-step later
        if (
            self.render_origin is RenderOrigin.PIPELINE_REPLICATION
            and (pipeline_name := os.environ.get('PIPELINE_NAME'))
        ):
            pipeline_metadata['replication_pipeline_name'] = pipeline_name

        if bg := effective_definition.get('background_image'):
            pipeline_metadata['background_image'] = bg

        for variant in pipeline_metadata.get('definition').variants():
            if not variant.has_main_repository():
                raise RuntimeError(
                    f"No main repository for pipeline definition {pipeline_definition.name}."
                )
            pipeline_metadata['pipeline_name'] = definition_descriptor.effective_pipeline_name()

        with makoutil.template_lock:
            template = mako.template.Template(template_contents, lookup=self.lookup) # nosec B702

        try:
            definition_descriptor.pipeline = template.render(
                    config_set=self.cfg_set,
                    pipeline=pipeline_metadata,
            )
        except:
            # use mako-provided function to render the stacktrace. This will properly translate
            # linenos and fill in information into the generated stacktrace
            rendered_stacktrace = mako.exceptions.text_error_template().render()
            raise RuntimeError(
                f"An exception occured when rendering pipeline '{pipeline_definition.name}': "
                f"{rendered_stacktrace}"
            ) from None

        return definition_descriptor


class RenderStatus(Enum):
    SUCCEEDED = 0
    FAILED = 1


@dataclasses.dataclass
class RenderResult:
    definition_descriptor: DefinitionDescriptor
    render_status: RenderStatus
    error_details: str = None
    exception: Exception = None


class DeployStatus(IntEnum):
    SUCCEEDED = 1
    FAILED = 2
    SKIPPED = 4
    CREATED = 8


@dataclasses.dataclass(frozen=True)
class DeployResult:
    definition_descriptor: DefinitionDescriptor
    deploy_status: DeployStatus
    error_details: str = None

    def ok(self):
        if self.deploy_status & DeployStatus.SUCCEEDED:
            return True
        elif self.deploy_status is DeployStatus.SKIPPED:
            return True
        elif self.deploy_status is DeployStatus.FAILED:
            return False
        else:
            raise NotImplementedError(self.deploy_status)


class DefinitionDeployer:
    def deploy(self, definition_descriptor):
        raise NotImplementedError('subclasses must overwrite')


class NoOpDeployer:
    def deploy(self, definition_descriptor):
        logger.info(f'Skipped deployment of pipeline {definition_descriptor.pipeline_name} (NoOp)')
        return DeployResult(
                definition_descriptor=definition_descriptor,
                deploy_status=DeployStatus.SUCCEEDED,
            )


class FilesystemDeployer(DefinitionDeployer):
    def __init__(self, base_dir):
        self.base_dir = existing_dir(base_dir)

    def deploy(self, definition_descriptor):
        try:
            with open(os.path.join(self.base_dir, definition_descriptor.pipeline_name), 'w') as f:
                f.write(definition_descriptor.pipeline)
            return DeployResult(
                definition_descriptor=definition_descriptor,
                deploy_status=DeployStatus.SUCCEEDED,
            )
        except Exception as e:
            logger.warning(e)
            return DeployResult(
                definition_descriptor=definition_descriptor,
                deploy_status=DeployStatus.FAILED,
            )


class ConcourseDeployer(DefinitionDeployer):
    def __init__(
        self,
        cfg_set,
        unpause_pipelines: bool,
        unpause_new_pipelines: bool=False,
        expose_pipelines: bool=True,
    ):
        self.cfg_set = cfg_set
        self.unpause_pipelines = unpause_pipelines
        self.unpause_new_pipelines = unpause_new_pipelines
        self.expose_pipelines = expose_pipelines

    def deploy(
        self,
        definition_descriptor: DefinitionDescriptor,
    ):
        pipeline_definition = definition_descriptor.pipeline
        pipeline_name = definition_descriptor.pipeline_name
        try:
            concourse_cfg = definition_descriptor.concourse_target_cfg

            api = ccc.concourse.client_from_cfg_name(
                concourse_cfg_name=concourse_cfg.name(),
                team_name=definition_descriptor.concourse_target_team,
            )

            try:
                response = api.set_pipeline(
                    name=pipeline_name,
                    pipeline_definition=pipeline_definition
                )
            except requests.exceptions.HTTPError as e:
                # We sometimes see this non-descript error. According to an old Concourse-Issue
                # this might happen due to concurrent saves. Wait a random time and try again.
                err_content = (
                    b'failed to save config: comparison with existing config failed during save'
                )
                if e.response.status_code == 500 and e.response.content == err_content:
                    time.sleep(random.randrange(5,30))
                    response = api.set_pipeline(
                        name=pipeline_name,
                        pipeline_definition=pipeline_definition
                    )
                else:
                    raise

            logger.info(
                'Deployed pipeline: ' + pipeline_name +
                ' to team: ' + definition_descriptor.concourse_target_team
            )

            SetPipelineResult = concourse.client.model.SetPipelineResult
            if self.unpause_pipelines:
                logger.info(f'Unpausing pipeline {pipeline_name}')
                api.unpause_pipeline(pipeline_name=pipeline_name)
            elif self.unpause_new_pipelines and response is SetPipelineResult.CREATED:
                logger.info(f'Unpausing new {pipeline_name=}')
                api.unpause_pipeline(pipeline_name=pipeline_name)

            if self.expose_pipelines:
                api.expose_pipeline(pipeline_name=pipeline_name)

            deploy_status = DeployStatus.SUCCEEDED
            if response is SetPipelineResult.CREATED:
                deploy_status |= DeployStatus.CREATED
            elif response is SetPipelineResult.UPDATED:
                pass
            else:
                raise NotImplementedError

            return DeployResult(
                definition_descriptor=definition_descriptor,
                deploy_status=deploy_status,
            )
        except Exception as e:
            import traceback
            traceback.print_exc()
            logger.warning(e)
            return DeployResult(
                definition_descriptor=definition_descriptor,
                deploy_status=DeployStatus.FAILED,
                error_details=traceback.format_exc(),
            )


class ReplicationResultProcessor:
    def __init__(
        self,
        cfg_set,
        unpause_new_pipelines: bool=True,
        remove_pipelines: bool=True,
        remove_pipelines_filter: typing.Callable[[str], bool]=None,
        reorder_pipelines: bool=True,
        job_mapping=None,
    ):
        '''
        @param remove_pipelines_filter: pipeline-names the filter matches are never removed
        '''
        self._cfg_set = cfg_set
        self._job_mapping = job_mapping
        self.unpause_new_pipelines = unpause_new_pipelines
        self.remove_pipelines = remove_pipelines
        self.remove_pipelines_filter = remove_pipelines_filter
        self.reorder_pipelines = reorder_pipelines

        CleanupPolicy = model.concourse.PipelineCleanupPolicy
        if self._job_mapping and \
                self._job_mapping.cleanup_policy() is CleanupPolicy.NO_CLEANUP:
            self.remove_pipelines = False
            logging.info(f'{job_mapping=} will not cleanup extra pipelines due to policy')

    def process_results(self, results):
        # collect pipelines by concourse target (concourse_cfg, team_name) as key
        concourse_target_results = {}
        for result in results:
            definition_descriptor = result.definition_descriptor
            concourse_target_key = definition_descriptor.concourse_target_key()
            if concourse_target_key not in concourse_target_results:
                concourse_target_results[concourse_target_key] = []
            concourse_target_results[concourse_target_key].append(result)

        # evaluate results
        failed_descriptors = [
            d for d in results
            if not d.deploy_status & DeployStatus.SUCCEEDED
        ]
        failed_count = len(failed_descriptors)

        if failed_count > 0:
            self.remove_pipelines = False
            logger.info(
                'Failures occurred during pipeline-replication. Will not cleanup pipelines.'
            )

        for concourse_target_key, concourse_results in concourse_target_results.items():
            # TODO: implement eq for concourse_cfg
            concourse_cfg, concourse_team = next(iter(
                concourse_results)).definition_descriptor.concourse_target()

            concourse_results = concourse_target_results[concourse_target_key]
            concourse_api = ccc.concourse.client_from_cfg_name(
                concourse_cfg_name=concourse_cfg.name(),
                team_name=concourse_team,
            )

            # find pipelines to remove
            if self.remove_pipelines:
                deployed_pipeline_names = set(map(
                    lambda r: r.definition_descriptor.pipeline_name, concourse_results
                ))

                pipelines_to_remove = set(concourse_api.pipelines()) - deployed_pipeline_names

                if self.remove_pipelines_filter:
                    logger.info(f'before applying filter: {pipelines_to_remove=}')
                    pipelines_to_remove = {
                        name for name in pipelines_to_remove
                        if not self.remove_pipelines_filter(name)
                    }
                    logger.info(f'after applying filter: {pipelines_to_remove=}')

                for pipeline_name in pipelines_to_remove:
                    logger.info('removing pipeline: {p}'.format(p=pipeline_name))
                    concourse_api.delete_pipeline(pipeline_name)

            # trigger resource checks in new pipelines
            self._initialise_new_pipeline_resources(concourse_api, concourse_results)
            if self.reorder_pipelines:
                # order pipelines alphabetically
                pipeline_names = list(concourse_api.pipelines())
                pipeline_names.sort()
                concourse_api.order_pipelines(pipeline_names)

        logger.info('Successfully replicated {d} pipeline(s)'.format(d=len(results) - failed_count))

        if failed_count == 0:
            return True

        logger.warning(f'Errors occurred whilst replicating pipeline(s): {failed_count=}')

        def should_notify_pipeline_owners(deploy_result: DeployResult):
            if deploy_result.deploy_status & DeployStatus.SUCCEEDED:
                # actually, this codepath should not be hit (we filter before-hand)
                pipeline_name = deploy_result.definition_descriptor.pipeline_name
                logger.warning(f'will not notify (no err): {pipeline_name=}')
                return False

            # if one of those exceptions was raised, presumably, there was either a
            # (hopefully) transient issue (e.g. network connectivity), or a programming error
            # in our template (in which case we should not bother end-users)
            ignore_exceptions = (
                ArithmeticError,
                AttributeError,
                BufferError,
                EOFError,
                ImportError,
                MemoryError,
                NameError,
                OSError,
                ReferenceError,
                RecursionError,
                SyntaxError,
                TypeError,
            )

            if type(deploy_result.definition_descriptor.exception) in ignore_exceptions:
                return False
            else:
                return True

        all_notifications_succeeded = True
        for failed_descriptor in failed_descriptors:
            logger.warning(failed_descriptor.definition_descriptor.pipeline_name)
            failed_descriptor: DeployResult
            # failed_descriptor has type DeployResult - XXX: should rename variable
            if not should_notify_pipeline_owners(
                deploy_result=failed_descriptor,
            ):
                logger.warning(
                    'will not notify (likely the error is not on user-side '
                    f'{failed_descriptor.pipeline_name=} {failed_descriptor.error_details=}'
                )
                continue
            try:
                self._notify_broken_definition_owners(failed_descriptor)
            except Exception:
                logger.warning('an error occurred whilst trying to send error notifications')
                traceback.print_exc()
                all_notifications_succeeded = False

        # signall error only if error notifications failed
        return all_notifications_succeeded

    def _notify_broken_definition_owners(self, failed_descriptor):
        definition_descriptor = failed_descriptor.definition_descriptor
        main_repo = definition_descriptor.main_repo
        repo_owner, repo_name = main_repo['path'].split('/')
        repo_url = urljoin(main_repo['hostname'], repo_owner, repo_name)
        github_cfg = ccc.github.github_cfg_for_repo_url(
            repo_url=repo_url,
            cfg_factory=self._cfg_set,
        )
        github_api = ccc.github.github_api(github_cfg)

        repo_helper = ccc.github.github_repo_helper(
            host=main_repo['hostname'],
            org=repo_owner,
            repo=repo_name,
            branch=main_repo['branch'],
        )

        recipients = set(github.codeowners.resolve_email_addresses(
            codeowners_entries=github.codeowners.enumerate_codeowners_from_remote_repo(
                repo=repo_helper.repository
            ),
            github_api=github_api,
        ))

        # in case no codeowners are available, resort to using the committer
        if not recipients:
            head_commit = repo_helper.repository.commit(main_repo['branch'])
            user_ids = {
                user_info.get('login')
                for user_info
                in (head_commit.committer, head_commit.author)
                if user_info and user_info.get('login')
            }
            for user_id in user_ids:
                user = github_api.user(user_id)
                if user.email:
                    recipients.add(user.email)

        # if there are still no recipients available print a warning
        if not recipients:
            logger.warning(textwrap.dedent(
                f"""
                Unable to determine recipient for pipeline '{definition_descriptor.pipeline_name}'
                found in branch '{main_repo['branch']}' ({main_repo['path']}). Please make sure that
                CODEOWNERS and committers have exposed a public e-mail address in their profile.
                """
            ))
        else:
            logger.info(f'Sending notification e-mail to {recipients} ({main_repo["path"]})')
            email_cfg = self._cfg_set.email()
            _send_mail(
                email_cfg=email_cfg,
                recipients=recipients,
                subject='Your pipeline definition in {repo} is erroneous'.format(
                    repo=main_repo['path'],
                ),
                mail_template=textwrap.dedent(
                f'''
                    The pipeline definition for {definition_descriptor.pipeline_name=}
                    on {main_repo["branch"]=} failed to be rendered.
                    Error details:
                    {str(failed_descriptor.error_details)}
                '''
                ),
            )

    def _initialise_new_pipeline_resources(
        self,
        concourse_api: concourse.client.api.ConcourseApiBase,
        results,
    ):
        newly_deployed_pipeline_names = [
            result.definition_descriptor.pipeline_name for result in results
            if result.deploy_status & DeployStatus.CREATED
        ]

        for pipeline_name in newly_deployed_pipeline_names:
            if self.unpause_new_pipelines:
                logger.info(f'unpausing new {pipeline_name=}')
                concourse_api.unpause_pipeline(pipeline_name)

            logger.info(f'triggering initial resource check for {pipeline_name=}')

            for pipeline_config_resource in concourse_api.pipeline_resources(pipeline_name):
                pipeline_config_resource: concourse.client.api.PipelineConfigResource
                concourse_api.trigger_resource_check(
                    pipeline_name=pipeline_name,
                    resource_name=pipeline_config_resource.name,
                )


class PipelineValidationError(Exception):
    pass


class PipelineValidationResultProcessor:
    def process_results(self, results: typing.Iterable[DeployResult]):
        failed_results = [
            r for r in results if not r.deploy_status is DeployStatus.SUCCEEDED
        ]
        if failed_results:
            raise PipelineValidationError(
                '\n'.join((
                    f'{r.definition_descriptor.pipeline_name}: {r.error_details}'
                    for r in failed_results)
                )
            )

        return True


class PipelineReplicator:
    def __init__(
            self,
            definition_enumerators,
            descriptor_preprocessor,
            definition_renderer,
            definition_deployer,
            result_processor=None,
        ):
        self.definition_enumerators = definition_enumerators
        self.descriptor_preprocessor = descriptor_preprocessor
        self.definition_renderer = definition_renderer
        self.definition_deployer = definition_deployer
        self.result_processor = result_processor

        # keep track of generated pipelines to detect conflicts
        self._pipeline_names_lock = threading.Lock()
        self._pipeline_names = set()

    def _pipeline_name_conflict(self, definition_descriptor:DefinitionDescriptor):
        with self._pipeline_names_lock:
            pipeline_name = definition_descriptor.pipeline_name
            if pipeline_name in self._pipeline_names:
                return True
            self._pipeline_names.add(pipeline_name)

    def _enumerate_definitions(self):
        for enumerator in self.definition_enumerators:
            yield from enumerator.enumerate_definition_descriptors()

    def _process_definition_descriptor(self, definition_descriptor):
        if definition_descriptor.exception:
            return DeployResult(
                definition_descriptor=definition_descriptor,
                deploy_status=DeployStatus.SKIPPED,
                error_details=definition_descriptor.exception,
            )

        preprocessed = self.descriptor_preprocessor.process_definition_descriptor(
                definition_descriptor
        )
        result = self.definition_renderer.render(preprocessed)

        if self._pipeline_name_conflict(
            definition_descriptor=result.definition_descriptor,
        ):
            # early exit upon pipeline name conflict
            pipeline_name = result.definition_descriptor.pipeline_name
            logger.warning(f'duplicate pipeline name: {pipeline_name}')
            return DeployResult(
                definition_descriptor=definition_descriptor,
                deploy_status=DeployStatus.SKIPPED,
                error_details=f'duplicate pipeline name: {pipeline_name}',
            )

        if result.render_status == RenderStatus.SUCCEEDED:
            deploy_result = self.definition_deployer.deploy(result.definition_descriptor)
        else:
            deploy_result = DeployResult(
                definition_descriptor=definition_descriptor,
                deploy_status=DeployStatus.SKIPPED,
                error_details=result.error_details,
            )
        return deploy_result

    def _replicate(self):
        executor = ThreadPoolExecutor(max_workers=16)
        yield from executor.map(
            self._process_definition_descriptor,
            self._enumerate_definitions(),
        )

    def replicate(self):
        results = [
            result for result in self._replicate()
        ]

        if self.result_processor:
            return self.result_processor.process_results(results)
        else:
            return results



import json
import logging
import subprocess

import dockerfile_parse

import ci.log
import model.container_registry as mc

ci.log.configure_default_logging()
logger = logging.getLogger('build-oci-image.step')


def write_docker_cfg(
    dockerfile_path: str,
    docker_cfg_path: str
):
    with open(dockerfile_path) as f:
        parser = dockerfile_parse.DockerfileParser(fileobj=f)
        relevant_image_refs = parser.parent_images

    # use dict to deduplicate by cfg-name (which we otherwise do not care about)
    container_registry_cfgs = {
        c.name(): c for c
        in (mc.find_config(img_ref) for img_ref in relevant_image_refs)
        if c is not None
    }.values()

    docker_cfg_auths = {}

    for container_registry_cfg in container_registry_cfgs:
        docker_cfg_auths.update(container_registry_cfg.as_docker_auths())

    docker_cfg = {'auths': docker_cfg_auths}

    with open(docker_cfg_path, 'w') as f:
        json.dump(docker_cfg, f)


def prepare_qemu_and_binfmt_misc():
    ## needs to be run once to allow for cross-platform executions/builds
    ## see: https://github.com/multiarch/qemu-user-static
    subprocess.run((
        'docker',
        'run',
        '--rm',
        '--privileged',
        'multiarch/qemu-user-static',
        '--reset',
        '--persistent', 'yes',
    ))



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


from model.base import (
    BasicCredentials,
    NamedModelElement,
)


class CheckmarxCredentials(BasicCredentials):
    def client_id(self):
        return self.raw.get('client_id')

    def client_secret(self):
        return self.raw.get('client_secret')

    def domain(self):
        return self.raw.get('domain')

    def scope(self):
        return self.raw.get('scope', 'sast_rest_api')

    def qualified_username(self):
        return f'{self.domain()}\\{self.username()}'


class CheckmarxConfig(NamedModelElement):
    def base_url(self):
        return self.raw.get('base_url')

    def team_id(self):
        return self.raw.get('team_id')

    def credentials(self):
        return CheckmarxCredentials(self.raw.get('credentials'))

    def _required_attributes(self):
        return 'credentials', 'base_url'



import datetime
import logging
import time

import checkmarx.client
import checkmarx.model as model
import checkmarx.util
import cnudie.iter
import ocm
import github.compliance.model as gcm

logger = logging.getLogger(__name__)


def _create_or_get_project(
    client: checkmarx.client.CheckmarxClient,
    name: str,
    team_id: str,
    is_public: bool = True,
):
    try:
        project_id = client.get_project_id_by_name(project_name=name, team_id=team_id)
        logger.info(f'Use existing Checkmarx project: {name}')
        return project_id
    except checkmarx.client.CXNotOkayException as e:
        if e.res.status_code == 404:
            logger.info(f'Create Checkmarx project: {name}')
            return client.create_project(name, team_id, is_public).json().get('id')
        else:
            raise e


class CheckmarxProject:
    def __init__(
        self,
        artifact_name: str,
        checkmarx_client: checkmarx.client.CheckmarxClient,
        project_details: model.ProjectDetails,
    ):
        self.artifact_name = artifact_name
        self.client = checkmarx_client
        self.project_details = project_details

    def poll_and_retrieve_scan(
        self,
        scan_id: int,
        component: ocm.Component,
        source: ocm.Source,
        timeout_seconds: int,
    ) -> model.ScanResult:

        def fetch_response(
                scan_id: int,
                timeout_seconds: int
            ) -> tuple[model.ScanResponse, gcm.ScanState, str, model.ScanStatistic]:
            clogger = checkmarx.util.component_logger(artifact_name=self.artifact_name)
            try:
                scan_response = self._poll_scan(
                    scan_id=scan_id,
                    timeout_seconds=timeout_seconds,
                )
            except Exception as e:
                clogger.error(f'scan with {scan_id=} failed: {e}')
                return None, gcm.ScanState.FAILED, None, None

            if scan_response.status_value() is not model.ScanStatusValues.FINISHED:
                clogger.error(
                    f'Scan of artifact "{self.artifact_name}:{source.version}" '
                    f'finished failed with {scan_response.status=}'
                )
                return scan_response, gcm.ScanState.FAILED, None, None

            scan_state = gcm.ScanState.SUCCEEDED
            report_url = self.client.routes.web_ui_scan_viewer(
                scan_id=scan_response.id,
                project_id=self.project_details.id,
            )

            clogger.info('scan finished. Retrieving scan statistics')
            statistics = self.scan_statistics(scan_id=scan_response.id)
            return scan_response, scan_state, report_url, statistics

        scan_response, scan_state, report_url, statistics = fetch_response(
            scan_id=scan_id,
            timeout_seconds=timeout_seconds,
        )

        return model.ScanResult(
            state=scan_state,
            scanned_element=cnudie.iter.SourceNode(
                path=(cnudie.iter.NodePathEntry(component),),
                source=source,
            ),
            project_id=self.project_details.id,
            artifact_name=self.artifact_name,
            scan_response=scan_response,
            scan_statistic=statistics,
            report_url=report_url,
            overview_url=self.client.routes.web_ui_scan_history(project_id=self.project_details.id),
        )

    def update_remote_project(self):
        self.client.update_project(self.project_details)

    def start_scan(self, scan_settings: model.ScanSettings):
        return self.client.start_scan(scan_settings)

    def _poll_scan(
        self,
        scan_id: int,
        polling_interval_seconds=60,
        timeout_seconds: int=3600,
    ) -> model.ScanResponse:
        def scan_finished():
            scan = self.client.get_scan_state(scan_id=scan_id)
            clogger = checkmarx.util.component_logger(artifact_name=self.artifact_name)
            clogger.info(f'polling for {scan_id=}. {scan.status.name=}, {scan.status.details.stage=}'
                f'{scan.status.details.step=}')
            if self.is_scan_finished(scan):
                return scan
            return False

        tm_start = datetime.datetime.now()
        tm_now = tm_start
        result = scan_finished()
        while not result and (tm_now - tm_start).total_seconds() < timeout_seconds:
            # keep polling until result is ready
            time.sleep(polling_interval_seconds)
            result = scan_finished()
            tm_now = datetime.datetime.now()
        if (tm_now - tm_start).total_seconds() >= timeout_seconds:
            scan = self.client.get_scan_state(scan_id=scan_id)
            raise RuntimeError(f'Scan of artifact "{scan.status.name=}", '
                f'{scan_id} aborted after timeout {timeout_seconds}s')

        return result

    def scan_statistics(self, scan_id: int):
        return self.client.get_scan_statistics(scan_id=scan_id)

    def get_last_scans(self):
        return self.client.get_last_scans_of_project(project_id=self.project_details.id)

    def is_scan_finished(self, scan: model.ScanResponse):
        if model.ScanStatusValues(scan.status.id) in (
                model.ScanStatusValues.FINISHED,
                model.ScanStatusValues.FAILED,
                model.ScanStatusValues.CANCELED,
        ):
            return True
        elif model.ScanStatusValues(scan.status.id) in (
            model.ScanStatusValues.NEW,
            model.ScanStatusValues.PRE_SCAN,
            model.ScanStatusValues.QUEUED,
            model.ScanStatusValues.SCANNING,
            model.ScanStatusValues.POST_SCAN,
            model.ScanStatusValues.SOURCE_PULLING_AND_DEPLOYMENT,
        ):
            return False
        else:
            raise NotImplementedError

    def is_scan_necessary(
        self,
        hash: str,
    ):
        remote_hash = self.project_details.get_custom_field(
            model.CustomFieldKeys.HASH,
        )
        if remote_hash != hash:
            return True
        else:
            return False

    def upload_zip(self, file, raise_on_error: bool = True):
        res = self.client.upload_zipped_source_code(self.project_details.id, file)
        if raise_on_error:
            res.raise_for_status()
        return res


def get_project_name(source_name: str):
    return source_name.replace('/', '_')


def init_checkmarx_project(
    checkmarx_client: checkmarx.client.CheckmarxClient,
    source_name: str,
    team_id: str,
) -> CheckmarxProject:

    project_name = get_project_name(source_name)

    project_id = _create_or_get_project(
        client=checkmarx_client,
        name=project_name,
        team_id=team_id,
    )

    project_details = checkmarx_client.get_project_by_id(project_id=project_id)

    return CheckmarxProject(
        checkmarx_client=checkmarx_client,
        artifact_name=source_name,
        project_details=project_details,
    )



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


from model.base import (
    NamedModelElement,
)


class SlackConfig(NamedModelElement):
    def api_token(self):
        return self.raw.get('api_token')

    def signing_secret(self):
        return self.raw.get('signing_secret')

    def _required_attributes(self):
        return ['api_token']



# SPDX-FileCopyrightText: 2021 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0
import ocm

import ctt.uploaders as uploaders


def test_labels_with_migration_hint_doesnt_overwrite_original_ref():
    firstLabel = ocm.Label(
        name='another_label',
        value='test'
    )
    secondLabel = ocm.Label(
        name=uploaders.original_ref_label_name,
        value='original-repo-0.com/my-image:1.0.0'
    )

    res = ocm.Resource(
        name='my-image',
        version='1.0.0',
        type=ocm.ArtefactType.OCI_IMAGE,
        access=ocm.OciAccess(
            imageReference='target-repo.com/my-image:1.0.0',
        ),
        labels=[
            firstLabel,
            secondLabel,
        ]
    )

    labels = uploaders.labels_with_migration_hint(res, 'original-repo-1.com/my-image:1.0.0')

    assert len(labels) == 2
    assert labels[0] == firstLabel
    assert labels[1] == secondLabel


def test_labels_with_migration_hint_adds_original_ref_if_not_present():
    firstLabel = ocm.Label(
        name='another_label',
        value='test'
    )
    expectedOriginalRefLabel = ocm.Label(
        name=uploaders.original_ref_label_name,
        value='original-repo.com/my-image:1.0.0'
    )

    res = ocm.Resource(
        name='my-image',
        version='1.0.0',
        type=ocm.ArtefactType.OCI_IMAGE,
        access=ocm.OciAccess(
            imageReference='target-repo.com/my-image:1.0.0',
        ),
        labels=[
            firstLabel,
        ]
    )

    labels = uploaders.labels_with_migration_hint(res, 'original-repo.com/my-image:1.0.0')

    assert len(labels) == 2
    assert labels[0] == firstLabel
    assert labels[1] == expectedOriginalRefLabel



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import collections.abc
import enum

from model.base import (
    ModelBase,
    NamedModelElement,
)


class DeliveryConfig(NamedModelElement):
    '''
    Not intended to be instantiated by users of this module
    '''
    def auth(self):
        return Auth(self.raw.get('auth'))

    def service(self):
        return DeliverySvcCfg(self.raw.get('service'))

    def dashboard(self):
        return DeliveryDashboardCfg(self.raw.get('dashboard'))

    def db_cfg_name(self):
        return self.raw.get('db_cfg_name')

    def _optional_attributes(self):
        yield from super()._optional_attributes()
        yield from [
            'db_cfg_name',
        ]

    def _required_attributes(self):
        yield from super()._required_attributes()
        yield from [
            'service',
        ]


class OAuthType(enum.Enum):
    GITHUB = 'github'


class Auth(ModelBase):
    def oauth_cfgs(self):
        return [OAuth(raw) for raw in self.raw.get('oauth_cfgs')]

    def oauth_cfg(self, github_cfg):
        for oc in self.oauth_cfgs():
            if oc.github_cfg() == github_cfg:
                return oc
        raise KeyError(f'no oauth cfg for {github_cfg}')


class OAuth(ModelBase):
    def name(self) -> str:
        return self.raw.get('name')

    def type(self):
        return OAuthType(self.raw.get('type'))

    def github_cfg(self):
        return self.raw.get('github_cfg')

    def oauth_url(self):
        return self.raw.get('oauth_url')

    def token_url(self):
        return self.raw.get('token_url')

    def client_id(self):
        return self.raw.get('client_id')

    def client_secret(self):
        return self.raw.get('client_secret')

    def scope(self):
        return self.raw.get('scope')


class SigningCfg(ModelBase):
    def id(self) -> str:
        return self.raw.get('id')

    def algorithm(self):
        return self.raw.get('algorithm', 'HS256')

    def secret(self):
        return self.raw.get('secret')

    def public_key(self):
        return self.raw.get('public_key')

    def purpose_labels(self) -> list[str]:
        return self.raw.get('purpose_labels', [])


class DeliveryDashboardCfg(ModelBase):
    def deployment_name(self):
        return self.raw.get('deployment_name', 'delivery-dashboard')


class DeliverySvcCfg(ModelBase):
    def deployment_name(self):
        return self.raw.get('deployment_name', 'delivery-service')

    def signing_cfgs(
        self,
        purpose_label: str = None,
    ) -> collections.abc.Generator[SigningCfg, None, None]:
        cfgs = self.raw.get('signing')

        if not purpose_label:
            yield from [
                SigningCfg(raw) for raw in cfgs
            ]
            return

        for raw_signing_cfg in cfgs:
            signing_cfg = SigningCfg(raw_signing_cfg)

            if purpose_label in signing_cfg.purpose_labels():
                yield signing_cfg


class DeliveryEndpointsCfg(NamedModelElement):
    def service_host(self):
        return self.raw['service_host']

    def base_url(self):
        return f'http://{self.service_host()}'

    def dashboard_host(self):
        return self.raw['dashboard_host']

    def dashboard_url(self):
        return f'https://{self.dashboard_host()}'



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import enum
import dataclasses
import textwrap
import typing

import dacite

import version
import ci.util
from concourse.model.job import (
    JobVariant,
)
from concourse.model.step import (
    PipelineStep,
    PrivilegeMode,
    PullRequestNotificationPolicy,
)
from concourse.model.base import (
  AttributeSpec,
  EnumWithDocumentation,
  EnumValueWithDocumentation,
  Trait,
  TraitTransformer,
  ScriptType,
)
from model.base import(
    ModelValidationError,
)
import concourse.model.traits.images

OciImageCfg = concourse.model.traits.images.OciImageCfg


class NextVersion(EnumWithDocumentation):
    BUMP_MAJOR = EnumValueWithDocumentation(
        value='bump_major',
        doc='Increments the major version of the next development cycle',
    )

    BUMP_MINOR = EnumValueWithDocumentation(
        value='bump_minor',
        doc='Increments the minor version of the next development cycle',
    )

    BUMP_PATCH = EnumValueWithDocumentation(
        value='bump_patch',
        doc='Increments the patch version of the next development cycle',
    )

    NOOP = EnumValueWithDocumentation(
        value='noop',
        doc='No change to the next development cycle version done',
    )


class ReleaseNotesPolicy(EnumWithDocumentation):
    DEFAULT = EnumValueWithDocumentation(
        value='default',
        doc='Create release notes and add them to the GitHub release.',
    )
    DISABLED = EnumValueWithDocumentation(
        value='disabled',
        doc='Do not create release notes.',
    )


class ReleaseCommitPublishingPolicy(EnumWithDocumentation):
    TAG_AND_PUSH_TO_BRANCH = EnumValueWithDocumentation(
        value='push_to_branch',
        doc='publish release tag to branch',
    )

    TAG_ONLY = EnumValueWithDocumentation(
        value='tag_only',
        doc='publish release tag to dead-end',
    )

    TAG_AND_MERGE_BACK = EnumValueWithDocumentation(
        value='tag_and_merge_back',
        doc='publish release tag to dead-end and merge back release commit to default branch',
    )


class TagConflictAction(enum.StrEnum):
    IGNORE = 'ignore'
    FAIL = 'fail'
    INCREMENT_PATCH_VERSION = 'increment-patch-version'


@dataclasses.dataclass(kw_only=True)
class Asset:
    '''
    base class for release assets. Not intended to be instantiated

    Assets are references to data from build (build log, files, or directory trees) that are
    to be included in Component-Descriptor as resources. Each asset is added as one resource.

    Parameters for OCM Resource:

    - ocm_labels: OCM-Labels to be added to resource
    - name: resource name
    - artefact_type: resource type
    - purposes: value for label `gardener.cloud/purposes`

    Other Parameters:

    - step_name: pipelinestep-name of data-source
    - comment: unused
    - type: asset-type - overwritten by subclasses

    Purpose-labels will be added as OCM-Label `gardener.cloud/purposes`.
    They are used to identify (in a machine-readable manner) the purpose / semantics
    of the included assets. Values are not (yet) standardised. It is recommended to use the
    following labels (multiple may be specified):

    build
    codecoverage
    integrationtest
    linter
    sast          static code analysis results
    test
    unittest
    '''
    ocm_labels: list[dict[str, object]] = dataclasses.field(default_factory=list)
    type: str = None # must overwrite
    name: str = None
    step_name: str
    artefact_type: str = 'application/data'
    purposes: list[str] = dataclasses.field(default_factory=list)
    comment: str | None = None


class FileAssetMode(enum.StrEnum):
    TAR = 'tar'
    SINGLE_FILE = 'single-file'


@dataclasses.dataclass(kw_only=True)
class BuildstepFileAsset(Asset):
    '''
    A reference to files from a build-step to be added to OCM Component Descriptor as resource.

    `step_output_dir` specifies the name of the output directory from which to read specified files.
    It must be declared from the build-step (step_name parameter) using the `output_dir`
    parameter.

    `path` is interpreted relative to `step_output_dir`. Using Unix globbing, any matching files
    will be included (use `*` to include all, except files starting with period (`.`) character).

    `prefix` will be removed from fnames in tar (similar to tar's -C option). Only effective in
    TAR mode (ignored otherwise). If any path does not start w/ prefix, this is not considered an
    error. Instead, such paths will silently be kept unchanged.

    By default, matching files will be wrapped in a compressed TAR archive (controlled by `mode`
    attribute). If mode is set to `single-file`, `path` must match exactly one regular file.

    In either case, it is considered an error if no matching files are found.
    '''
    type: str = 'build-step-file'
    step_output_dir: str
    path: str # rel-path; globbing is allowed
    prefix: str | None = None
    mode: FileAssetMode = FileAssetMode.TAR


@dataclasses.dataclass(kw_only=True)
class BuildstepLogAsset(Asset):
    '''
    An (additional) release asset to be included as part of a release. Hardcoded for the
    special-case of including build-step-logs as resource in released component-descriptor
    with a srcRef to main repository (all of which might be made more flexible if needed).
    '''
    type: str = 'build-step-log'
    artefact_type = 'text/plain'

    def __post_init__(self):
        if not self.name:
            self.name = f'{self.step_name}-build-step'

        if self.purposes:
            self.ocm_labels.append({
                'name': 'gardener.cloud/purposes',
                'value': self.purposes,
            })

        if self.comment:
            self.ocm_labels.append({
                'name': 'gardener.cloud/comment',
                'value': self.comment,
            })


ATTRIBUTES = (
    AttributeSpec.optional(
        name='assets',
        default=[],
        doc='''
        additional release-assets to publish.
        ''',
        type=list[BuildstepLogAsset | BuildstepFileAsset, ...],
    ),
    AttributeSpec.optional(
        name='nextversion',
        default=NextVersion.BUMP_MINOR.value,
        doc='specifies how the next development version is to be calculated',
        type=NextVersion,
    ),
    AttributeSpec.optional(
        name='release_callback',
        default=None,
        doc='''
        an optional callback that is called during release commit creation. The callback is passed
        the absolute path to the main repository's work tree via environment variable `REPO_DIR`.
        Any changes left inside the worktree are added to the resulting release commit.
        ''',
    ),
    AttributeSpec.optional(
        name='release_callback_image_reference',
        default=None,
        doc='''
        if specified, the release_callback will be run in a virtualisation container using the
        chosen container image (if not specified, the callback is run as a subprocess)
        ''',
        type=OciImageCfg,
    ),
    AttributeSpec.optional(
        name='rebase_before_release',
        default=False,
        doc='''
        whether or not a rebase against latest branch head should be done before publishing
        release commits.
        ''',
        type=bool,
    ),
    AttributeSpec.optional(
        name='next_version_callback',
        default=None,
        doc='''
        an optional callback that is called during next version commit creation.
        The callback is passed the absolute path to the main repository's work tree via environment
        variable `REPO_DIR`.
        Any changes left inside the worktree are added to the commit bumping the version
        immediately after the release-commit.
        ''',
    ),
    AttributeSpec.optional(
        name='release_notes_policy',
        default=ReleaseNotesPolicy.DEFAULT.value,
        doc='''
        configures the release notes handling policy
        ''',
        type=ReleaseNotesPolicy,
    ),
    AttributeSpec.optional(
        name='release_commit_publishing_policy',
        default=ReleaseCommitPublishingPolicy.TAG_AND_PUSH_TO_BRANCH,
        doc='''
        configures how the release commit should be published
        ''',
        type=ReleaseCommitPublishingPolicy,
    ),
    AttributeSpec.optional(
        name='commit_message_prefix',
        default=None,
        doc='''
        an optional prefix for release commit messages
        ''',
        type=str,
    ),
    AttributeSpec.optional(
        name='next_version_commit_message_prefix',
        default=None,
        doc='''
        an optional prefix for the commit message of the commit bumping the release version
        immediately after the release-commit
        ''',
        type=str,
    ),
    AttributeSpec.optional(
        name='merge_release_to_default_branch_commit_message_prefix',
        default=None,
        doc='''
        an optional prefix for the merge-commit message when merging back release-commit from tag
        to default branch
        ''',
        type=str,
    ),
    AttributeSpec.optional(
        name='git_tags',
        default=[{'ref_template': 'refs/tags/{VERSION}'}],
        doc='''
        a list of tags to tag the release commit with, **at least one**. The following
        placeholders are available:

        - {VERSION}: The version to be released (i.e. the 'effective version')

        The first tag will be used to create the GitHub-release.
        ''',
        type=list
    ),
    AttributeSpec.optional(
        name='release_on_github',
        default=True,
        doc='''
        if true, a github release is published.
        ''',
        type=bool
    ),
    AttributeSpec.optional(
        name='on_tag_conflict',
        default=TagConflictAction.IGNORE,
        doc=textwrap.dedent('''\
            specifies the action to take if the tag to be pushed (`refs/tags/<effective-version>`)
            already exists (relevant, if `release_on_github` is not set to `False`).
            Such cases can occur if pushing of "bump-commit" after previous release failed.
            Default value is chosen for backwards-compatibility.
            `fail` will lead to the `version` step to fail (which will shorten the time to
            discover this error, and thus save time)
            `increment-patch-version` will increment effective version's patchlevel (and thus
            avoid a conflict)
        '''),
        type=TagConflictAction,
    ),
)


class ReleaseTrait(Trait):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def _attribute_specs(cls):
        return ATTRIBUTES

    @property
    def assets(self) -> list[BuildstepLogAsset | BuildstepFileAsset, ...]:
        if not (raw_assets := self.raw.get('assets')):
            return []

        def deserialise(raw_asset: dict):
            if not 'type' in raw_asset:
                raise ModelValidationError(
                    f'`type` must be present for {raw_asset=}',
                )

            if (typename := raw_asset['type']) == BuildstepFileAsset.type:
                data_class = BuildstepFileAsset
            elif typename == BuildstepLogAsset.type:
                data_class = BuildstepLogAsset
            else:
                raise ModelValidationError(typename)

            return dacite.from_dict(
                data_class=data_class,
                data=raw_asset,
                config=dacite.Config(
                    cast=(enum.Enum,),
                ),
            )

        return [
            deserialise(raw_asset)
            for raw_asset in raw_assets
        ]

    def nextversion(self):
        return self.raw['nextversion']

    def release_callback_path(self):
        return self.raw['release_callback']

    def release_callback_image_reference(self) -> typing.Optional[OciImageCfg]:
        if not (raw := self.raw.get('release_callback_image_reference')):
            return None

        return OciImageCfg(raw_dict=raw)

    def next_version_callback_path(self):
        return self.raw['next_version_callback']

    def rebase_before_release(self):
        return self.raw['rebase_before_release']

    def release_notes_policy(self) -> ReleaseNotesPolicy:
        return ReleaseNotesPolicy(self.raw.get('release_notes_policy'))

    def release_commit_publishing_policy(self) -> ReleaseCommitPublishingPolicy:
        return ReleaseCommitPublishingPolicy(self.raw['release_commit_publishing_policy'])

    def release_commit_message_prefix(self) -> str:
        return self.raw.get('commit_message_prefix')

    def next_cycle_commit_message_prefix(self) -> str:
        return self.raw.get('next_version_commit_message_prefix')

    def merge_release_to_default_branch_commit_message_prefix(self) -> str:
        return self.raw.get('merge_release_to_default_branch_commit_message_prefix')

    def git_tags(self):
        '''
        all tags to be created in addition to the "github-release-tag" (without the gh-release-tag)
        '''
        return self.raw.get('git_tags')[1:]

    def github_release_tag(self):
        if tags := self.raw.get('git_tags'):
            return tags[0]
        return None

    def release_on_github(self) -> bool:
        return self.raw['release_on_github']

    @property
    def on_tag_conflict(self) -> TagConflictAction:
        return TagConflictAction(self.raw['on_tag_conflict'])

    def validate(self):
        super().validate()
        if self.nextversion() == version.NOOP and self.next_version_callback_path():
            raise ModelValidationError(
                f'not possible to configure "next_version_callback" if version is "{version.NOOP}"'
            )
        if not self.github_release_tag():
            raise ModelValidationError('At least one tag must be configured for the release.')

        # ensure the form of the first tag is as expected - otherwise no release can be created
        if not self.github_release_tag()['ref_template'].startswith('refs/tags/'):
            raise ModelValidationError(
                "The first release-tag must be of the form 'refs/tags/<tagname>'."
            )

    # by default, all Trait merge list-arg-defaults. Disable to support overriding the default
    # release-tag
    def _apply_defaults(self, raw_dict):
        self.raw = ci.util.merge_dicts(
            self._defaults_dict(),
            raw_dict,
            list_semantics=None,
        )

    def transformer(self):
        return ReleaseTraitTransformer(trait=self)


class ReleaseTraitTransformer(TraitTransformer):
    name = 'release'

    def __init__(self, trait: ReleaseTrait, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.trait = trait

    def inject_steps(self):
        if self.trait.release_callback_image_reference():
            # we need privileged container in order to run callback in container
            privilege_mode = PrivilegeMode.PRIVILEGED
        else:
            privilege_mode = PrivilegeMode.UNPRIVILEGED

        # inject 'release' step
        self.release_step = PipelineStep(
            name='release',
            raw_dict={
                'privilege_mode': privilege_mode,
            },
            is_synthetic=True,
            pull_request_notification_policy=PullRequestNotificationPolicy.NO_NOTIFICATION,
            injecting_trait_name=self.name,
            script_type=ScriptType.PYTHON3,
        )
        self.release_step.set_timeout('2h')
        yield self.release_step

    def process_pipeline_args(self, pipeline_args: JobVariant):
        # we depend on all other steps
        for step in pipeline_args.steps():
            self.release_step._add_dependency(step)

            if step.name == 'helmcharts':
                self.release_step.add_input('helmcharts', 'helmcharts')

        # a 'release job' should only be triggered automatically if explicitly configured
        main_repo = pipeline_args.main_repository()
        if main_repo:
            if 'trigger' not in pipeline_args.raw['repo']:
                main_repo._trigger = False

        # validate assets if present
        for asset in self.trait.assets:
            if not pipeline_args.has_step(asset.step_name):
                raise ValueError(f'{asset=}\'s step_name refers to an absent build-step')

            if isinstance(asset, BuildstepFileAsset):
                asset: BuildstepFileAsset
                step = pipeline_args.step(asset.step_name)

                if not asset.step_output_dir in step.outputs().values():
                  raise ModelValidationError(
                    f'{step.name=} does not declare {asset.step_output_dir=}',
                  )

                self.release_step.add_input(
                  name=asset.step_output_dir,
                  variable_name=asset.step_output_dir,
                )

    @classmethod
    def dependencies(cls):
        return {'version', 'component_descriptor'}

    @classmethod
    def order_dependencies(cls):
        return {'publish'}



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import copy
import enum
import graphlib
import typing

from concourse.model.base import (
    ModelBase,
    select_attr,
)
from ci.util import not_none
from concourse.model.resources import RepositoryConfig
import concourse.model.step


_not_set = object() # sentinel


class AbortObsoleteJobs(enum.Enum):
    ALWAYS = 'always'
    ON_FORCE_PUSH_ONLY = 'on_force_push_only'
    NEVER = 'never'


class JobVariant(ModelBase):
    def __init__(self, name: str, raw_dict: dict, resource_registry, *args, **kwargs):
        self._main_repository_name = None
        self._resource_registry = not_none(resource_registry)
        self.variant_name = name
        self._publish_repos_dict = {}
        self._repos_dict = {}
        self._traits_dict = {}
        self._steps_dict = {}
        super().__init__(raw_dict=raw_dict, *args, **kwargs)

    @classmethod
    def _attribute_specs(cls):
        return ()

    def _known_attributes(self):
        return {
            'abort_obsolete_jobs',
            'repo',
            'repos',
            'steps',
            'traits',
        }

    def _children(self):
        yield from self.steps()
        yield from self.traits().values()
        yield from self.repositories()

    def traits(self):
        return self._traits_dict

    def trait(self, name, default=_not_set):
        trait =  self._traits_dict.get(name, default)
        if trait is _not_set:
            raise KeyError(f'trait {name=} not present for {self.job_name()=}')
        return trait

    def has_trait(self, name):
        return name in self.traits()

    def job_name(self):
        parts = []
        if self.has_main_repository():
            parts.append(self.main_repository().branch())
        parts.append(self.variant_name)
        parts.append('job')

        return '-'.join(parts)

    def steps(self) -> typing.Iterable[concourse.model.step.PipelineStep]:
        return self._steps_dict.values()

    def step_names(self):
        return map(select_attr('name'), self.steps())

    def _step_depends_on(
        self,
        step_name,
        step_dependency_name,
        dependencies,
        visited_steps=None
    ):
        '''return whether `step_name` depends on `step_dependency_name` given the passed
        dependencies-dictionary.

        Only considers steps defined by users.
        '''
        if not visited_steps:
            visited_steps = set()

        if step_name in visited_steps:
            return False

        visited_steps.add(step_name)

        if self.step(step_name).is_synthetic:
            return False
        # if step is not present in dependencies-dict, it has no relevant dependencies
        if not step_name in dependencies:
            return False
        if step_dependency_name in dependencies[step_name]:
            return True

        return any([
            self._step_depends_on(s, step_dependency_name, dependencies, visited_steps)
            for s in dependencies[step_name]
        ])

    def _find_and_resolve_publish_trait_circular_dependencies(
        self,
        dependencies: dict[str, set[str]],
        cycle_info: typing.Sequence[str],
    ):
        '''
        @param dependencies: dict of {step_name: dependencies} as understood by TopologicalSorter
        @cycle_info: sequence of step-names with circular dependency
        '''
        # handle circular dependencies caused by depending on the publish step, e.g.:
        # test -> publish -> ... -> prepare -> test
        updated_dependencies = copy.deepcopy(dependencies)
        custom_steps_depending_on_publish = [
            step_name for step_name in cycle_info
            if self._step_depends_on(
                step_name=step_name,
                step_dependency_name='publish',
                dependencies=updated_dependencies,
            )
        ]

        if 'prepare' in cycle_info and custom_steps_depending_on_publish:
            for step_name in custom_steps_depending_on_publish:
                if step_name in updated_dependencies['prepare']:
                    updated_dependencies['prepare'].remove(step_name)

        return updated_dependencies

    def _find_and_resolve_release_trait_circular_dependencies(
        self,
        dependencies,
        cycle_info: typing.Sequence[str],
    ):
        updated_dependencies = copy.deepcopy(dependencies)
        if (
            len(cycle_info) == 3
            and len(unique_steps := set(cycle_info)) == 2
            and 'release' in unique_steps # be defensive
        ):
            # We have a cycle that is represented as step -> release -> step. These
            # steps can actually be resolved easily by removing the dependency of the release-step
            # as they are only possible if the pipeline definition explicitly specifies
            # a step to depend on the release-trait.
            first_name, second_name = unique_steps
            if first_name == 'release':
                updated_dependencies[first_name].remove(second_name)
            else:
                # 'release' was the second entry
                updated_dependencies[second_name].remove(first_name)
        else:
            for step_name in cycle_info:
                step = self.step(step_name)
                if not step.is_synthetic:
                    continue # only patch away synthetic steps' dependencies

                for step_dependency_name in step._dependencies():
                    step_dependency = self.step(step_dependency_name)
                    if step_dependency.is_synthetic:
                        continue # leave dependencies between synthetic steps
                    # patch out dependency from synthetic step to custom step
                    if step_dependency_name in updated_dependencies[step_name]:
                        updated_dependencies[step_name].remove(step_dependency_name)

        return updated_dependencies

    def ordered_steps(self) -> typing.Generator[tuple[str], None, None]:
        dependencies = {
            step.name: step.depends() for step in self.steps()
        }
        # add dependencies on trait-defined steps
        for step in self.steps():
            dependencies[step.name] |= {
                s.name for s in self.steps()
                if s.injecting_trait_name() in step.trait_depends()
            }

        def iter_results(toposorter: graphlib.TopologicalSorter):
            while toposorter.is_active():
                ready_tasks = tuple(toposorter.get_ready())
                toposorter.done(*ready_tasks)
                yield ready_tasks

        try:
            toposorter = graphlib.TopologicalSorter(graph=dependencies)
            toposorter.prepare()
            return iter_results(toposorter=toposorter)
        except graphlib.CycleError as ce:
            cycle_steps = ce.args[1] # contains a list of circular steps
            dependencies = self._find_and_resolve_publish_trait_circular_dependencies(
                dependencies,
                cycle_info=cycle_steps,
            )
            try:
                # check whether resolving the dependency between the publish trait has already
                # fixed the issue
                toposorter = graphlib.TopologicalSorter(graph=dependencies)
                toposorter.prepare()
                return iter_results(toposorter=toposorter)
            except graphlib.CycleError as ce:
                cycle_steps = ce.args[1] # contains a list of circular steps
                dependencies = self._find_and_resolve_release_trait_circular_dependencies(
                    dependencies,
                    cycle_info=cycle_steps,
                )
                # try again - if there is still a cyclic dependency, this is probably caused
                # by a user error - so let it propagate
                toposorter = graphlib.TopologicalSorter(graph=dependencies)
                toposorter.prepare()
                return iter_results(toposorter=toposorter)

    def add_step(self, step: concourse.model.step.PipelineStep): # noqa
        if self.has_step(step.name):
            raise ValueError('conflict: pipeline definition already contained step {s}'.format(
                s=step.name
            )
            )
        self._steps_dict[step.name] = step

    def step(self, name) -> concourse.model.step.PipelineStep:
        if not (step := self._steps_dict.get(name)):
            raise ValueError(f'no such step: {name=}')
        return step

    def has_step(self, step_name):
        return step_name in self.step_names()

    def pr_repository(self, name):
        pr_repo = self.repository(name)
        return RepositoryConfig(
            raw_dict=dict(pr_repo.raw),
            logical_name=name,
            qualifier='pr',
            is_pull_request=True
        )

    def repositories(self):
        # TODO: introduce a common base class for "input resources"
        # (where Github and PR are two examples, and "time" will be the third)
        return self._repos_dict.values()

    def repository_names(self):
        return self._repos_dict.keys()

    def repository(self, name):
        return self._repos_dict[name]

    def has_main_repository(self):
        return self._main_repository_name is not None

    def main_repository(self):
        return self.repository(self._main_repository_name)

    def publish_repositories(self):
        return self._publish_repos_dict.values()

    def publish_repository(self, name):
        return self._publish_repos_dict[name]

    def has_publish_repository(self, name):
        return name in self._publish_repos_dict

    def __repr__(self):
        return f'JobVariant: {self.variant_name}'



import datetime
import enum
import functools
import logging
import textwrap
import time
import typing
import urllib.parse

import cachetools
import github3
import github3.issues
import github3.issues.issue
import github3.repos

import ocm
import requests

import ccc.github
import checkmarx.model
import cfg_mgmt.model as cmm
import ci.util
import cnudie.util
import concourse.model.traits.image_scan as image_scan
import delivery.client
import delivery.model
import github.codeowners
import github.compliance.issue
import github.compliance.milestone as gcmi
import github.compliance.model as gcm
import github.retry
import github.user
import github.util
import model.delivery

logger = logging.getLogger(__name__)

_compliance_label_os_outdated = github.compliance.issue._label_os_outdated
_compliance_label_checkmarx = github.compliance.issue._label_checkmarx
_compliance_label_credentials_outdated = github.compliance.issue._label_outdated_credentials
_compliance_label_no_responsible = github.compliance.issue._label_no_responsible
_compliance_label_no_rule = github.compliance.issue._label_no_rule
_compliance_label_no_status = github.compliance.issue._label_no_status
_compliance_label_undefined_policy = github.compliance.issue._label_undefined_policy

_ctx_label_prefix = github.compliance.issue._label_prefix_ctx


@cachetools.cached(cache={})
@github.retry.retry_and_throttle
def _all_issues(
    repository,
):
    return set(repository.issues())


def _criticality_classification(cve_score: float) -> gcm.Severity:
    if not cve_score or cve_score <= 0:
        return None

    if cve_score < 4.0:
        return gcm.Severity.LOW
    if cve_score < 7.0:
        return gcm.Severity.MEDIUM
    if cve_score < 9.0:
        return gcm.Severity.HIGH
    if cve_score >= 9.0:
        return gcm.Severity.CRITICAL


def _delivery_dashboard_url(
    component: ocm.Component,
    base_url: str,
):
    url = ci.util.urljoin(
        base_url,
        '#/component'
    )

    query = urllib.parse.urlencode(
        query={
            'name': component.name,
            'version': component.version,
            'view': 'bom',
        }
    )

    return f'{url}?{query}'


def _pluralise(prefix: str, count: int):
    if count == 1:
        return prefix
    return f'{prefix}s'


def _artifact_url(
    component: ocm.Component,
    artifact: ocm.Artifact,
) -> str | None:
    access = artifact.access

    if isinstance(access, ocm.OciAccess):
        return access.imageReference

    elif isinstance(access, ocm.S3Access):
        return f'http://{access.bucketName}.s3.amazonaws.com/{access.objectKey}'

    elif isinstance(access, ocm.GithubAccess):
        return access.repoUrl

    elif isinstance(access, ocm.LocalBlobAccess):
        ocm_repo = component.current_ocm_repo
        image_reference = ocm_repo.component_oci_ref(component.name)
        return f'{image_reference}@{access.localReference}'

    elif isinstance(access, ocm.LocalBlobGlobalAccess):
        return access.ref

    else:
        raise NotImplementedError(artifact.access)


def _compliance_status_summary(
    component: ocm.Component,
    artifacts: typing.Sequence[ocm.Artifact],
    report_urls: tuple[str] | set[str],
    issue_description: str,
    issue_value: str,
):
    if isinstance(artifacts[0].type, enum.Enum):
        artifact_type = artifacts[0].type.value
    else:
        artifact_type = artifacts[0].type

    artifact_versions = ', '.join((r.version for r in artifacts))

    artifact_urls = ' '.join(
        url
        for artefact in artifacts
        if (
            url := _artifact_url(
                component=component,
                artifact=artefact,
            )
        )
    )

    report_urls = '\n- '.join(report_urls)

    summary = textwrap.dedent(f'''\
        # Compliance Status Summary

        |    |    |
        | -- | -- |
        | Component | {component.name} |
        | Component-Version | {component.version} |
        | Artifact  | {artifacts[0].name} |
        | {_pluralise('Artifact-Version', len(artifacts))}  | {artifact_versions} |
        | Artifact-Type | {artifact_type} |
        | URLs | {artifact_urls} |
        | {issue_description} | {issue_value} |
    ''')

    summary += textwrap.dedent(f'''
        The aforementioned {_pluralise(artifact_type, len(artifacts))} yielded findings
        relevant for future release decisions.

        For viewing detailed scan {_pluralise('report', len(artifacts))}, see the following
        {_pluralise('Scan Report', len(artifacts))}:
    ''')

    return summary + '- ' + report_urls


def _ocm_result_group_template_vars(
    result_group: gcm.ScanResultGroup,
    delivery_dashboard_url: str,
) -> dict:
    component = result_group.component
    artifact_name = result_group.artifact
    artifacts = [gcm.artifact_from_node(res.scanned_element) for res in result_group.results]

    artifact_versions = ', '.join((artifact.version for artifact in artifacts))
    artifact_types = ', '.join(set(
        (
            artifact.type.value
            if isinstance(artifact.type, enum.Enum)
            else artifact.type
            for artifact in artifacts
        )
    ))

    return {
        'component_name': component.name,
        'component_version': component.version,
        'resource_name': artifact_name, # TODO: to be removed at some point use artifact_name instead
        'resource_version': artifact_versions, # TODO: to be removed use artifact_version instead
        'resource_type': artifact_types,       # TODO: to be removed use artifact_type instead
        'artifact_name': artifact_name,
        'artifact_version': artifact_versions,
        'artifact_type': artifact_types,
        'delivery_dashboard_url': delivery_dashboard_url,
    }


def _os_info_template_vars(
    result_group: gcm.ScanResultGroup,
) -> dict:
    worst_result = result_group.worst_result
    worst_result: gcm.OsIdScanResult
    os_info = worst_result.os_id
    os_name_and_version = f'{os_info.ID}:{os_info.VERSION_ID}'
    component = result_group.component
    artifacts = [gcm.artifact_from_node(res.scanned_element) for res in result_group.results]

    return {
        'summary': _compliance_status_summary(
            component=component,
            artifacts=artifacts,
            issue_value=os_name_and_version,
            issue_description='Outdated OS-Version',
            report_urls=(),
        ),
    }


def _checkmarx_template_vars(
    result_group: gcm.ScanResultGroup,
) -> dict:

    def iter_report_urls():
        for r in results:
            name = f'{r.scanned_element.source.name}:{r.scanned_element.source.version}'
            yield f'[Assessments for {name}]({r.report_url})'
            yield f'[Summary for {name}]({r.overview_url})'

    results: tuple[checkmarx.model.ScanResult] = result_group.results_with_findings
    worst_result: checkmarx.model.ScanResult = result_group.worst_result
    stat = worst_result.scan_statistic
    summary_str = (f'Findings: High: {stat.highSeverity}, Medium: {stat.mediumSeverity}, '
        f'Low: {stat.lowSeverity}, Info: {stat.infoSeverity}')
    artifacts = [gcm.artifact_from_node(res.scanned_element) for res in result_group.results]
    component = result_group.component

    crit = (f'Risk: {worst_result.scan_response.scanRisk}, '
        f'Risk Severity: {worst_result.scan_response.scanRiskSeverity}')

    return {
        'summary': _compliance_status_summary(
            component=component,
            artifacts=artifacts,
            issue_value=summary_str,
            issue_description='Checkmarx Scan Summary',
            report_urls=tuple(iter_report_urls()),
        ),
        'criticality_classification': crit,
    }


def _cfg_policy_violation_template_vars(result_group: gcm.ScanResultGroup) -> dict:
    results: tuple[gcm.CfgScanResult] = result_group.results_with_findings
    result = results[0]

    if result.scanned_element.responsible:
        # remove foremost "@" to prevent notification mails
        responsibles = '<br/>'.join([
            f'{r.name.removeprefix("@")} ({r.type.value})'
            for r in result.scanned_element.responsible.responsibles
        ])
        responsibles_len = len(result.scanned_element.responsible.responsibles)

    else:
        responsibles = 'unknown'
        responsibles_len = 1

    element_storage = result.scanned_element.element_storage

    summary = textwrap.dedent(f'''\
        # Compliance Status Summary
        |    |    |
        | -- | -- |
        | Element Storage | [{element_storage}](https://{element_storage}) |
        | Element Type | {result.scanned_element.element_type} |
        | Element Name | {result.scanned_element.element_name} |
        | {_pluralise("Responsible", responsibles_len)} | {responsibles} |
    ''')

    return {
        'summary': summary
    }


def _template_vars(
    result_group: gcm.ScanResultGroup,
    delivery_dashboard_url: str='',
) -> dict:
    scanned_element = result_group.results[0].scanned_element
    issue_type = result_group.issue_type

    if gcm.is_ocm_artefact_node(scanned_element):
        template_variables = _ocm_result_group_template_vars(
            result_group=result_group,
            delivery_dashboard_url=delivery_dashboard_url,
        )

    elif isinstance(scanned_element, cmm.CfgElementStatusReport):
        template_variables = {
            'cfg_element_name': scanned_element.element_name,
            'cfg_element_type': scanned_element.element_type,
            'cfg_element_storage': scanned_element.element_storage,
            'cfg_element_qualified_name': scanned_element.name,
        }

    else:
        raise TypeError(result_group)

    if issue_type == _compliance_label_os_outdated:
        template_variables |= _os_info_template_vars(result_group)

    elif issue_type == _compliance_label_checkmarx:
        template_variables |= _checkmarx_template_vars(result_group)

    elif issue_type in (
        _compliance_label_credentials_outdated,
        _compliance_label_no_responsible,
        _compliance_label_no_rule,
        _compliance_label_no_status,
        _compliance_label_undefined_policy,
    ):
        template_variables |= _cfg_policy_violation_template_vars(result_group)

    else:
        raise NotImplementedError(issue_type)

    return template_variables


def _scanned_element_repository(
    scanned_element: gcm.Target,
) -> github3.repos.repo.Repository:
    if gcm.is_ocm_artefact_node(scanned_element):
        source = cnudie.util.main_source(component=scanned_element.component)

        if not source.access.type is ocm.AccessType.GITHUB:
            raise NotImplementedError(source)

        org = source.access.org_name()
        name = source.access.repository_name()
        gh_api = ccc.github.github_api(repo_url=source.access.repoUrl)

        return gh_api.repository(org, name)

    elif isinstance(scanned_element, cmm.CfgElementStatusReport):
        gh_api = ccc.github.github_api(repo_url=scanned_element.element_storage)

        parsed_url = ci.util.urlparse(scanned_element.element_storage)
        path = parsed_url.path.strip('/')
        org, repo = path.split('/')

        return gh_api.repository(org, repo)

    else:
        raise TypeError(scanned_element)


def _scanned_element_assignees(
    scanned_element: gcm.Target,
    delivery_svc_client: delivery.client.DeliveryServiceClient | None,
    repository: github3.repos.repo.Repository,
    gh_api: github3.GitHub | github3.GitHubEnterprise,
) -> tuple[set[str], set[delivery.model.Status]]:
    '''
    Determines assignees for scanned-element based on its type.
        ocm-node:
        retrieve component-responsibles via delivery-service

        cfg-element:
        resolve cfg responsible mapping to github users

    Assignees are returned as set of GitHub usernames.
    GitHub instance (for username determination) is taken from `repository`.
    '''

    def iter_gh_usernames_from_responsibles_mapping(
        gh_api: github3.GitHub | github3.GitHubEnterprise,
        responsibles_mapping: cmm.CfgResponsibleMapping,
    ) -> typing.Generator[github.codeowners.Username, None, None]:
        unique_usernames = set()
        for responsible in responsibles_mapping.responsibles:
            if responsible.type == cmm.CfgResponsibleType.EMAIL:
                for username in github.codeowners.usernames_from_email_address(
                    email_address=responsible.name,
                    gh_api=gh_api,
                ):
                    unique_usernames.add(username)
            elif responsible.type == cmm.CfgResponsibleType.GITHUB:
                parsed = github.codeowners.parse_codeowner_entry(responsible.name)
                for username in github.codeowners.resolve_usernames(
                    codeowners_entries=[parsed],
                    github_api=gh_api,
                ):
                    unique_usernames.add(username)
            else:
                logger.warning(f'unable to process {responsible.type=}')
        yield from unique_usernames

    assignees: set[str] = set()
    statuses: set[delivery.model.Status] = set()

    if gcm.is_ocm_artefact_node(scanned_element):
        if not delivery_svc_client:
            return assignees, statuses

        artifact = gcm.artifact_from_node(scanned_element)
        try:
            responsibles, statuses = delivery_svc_client.component_responsibles(
                component=scanned_element.component,
                artifact=artifact,
            )
            statuses = set(statuses)

            gh_users = delivery.client.github_users_from_responsibles(
                responsibles=responsibles,
                github_url=repository.url,
            )

            assignees = set(
                gh_user.username for gh_user in gh_users
                if github.user.is_user_active(
                    username=gh_user.username,
                    github=gh_api,
                )
            )

        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404:
                logger.warning(f'Delivery Service returned 404 for '
                    f'{scanned_element.component.name=}, {artifact.name=}')
                return assignees, statuses

            else:
                raise

    elif isinstance(scanned_element, cmm.CfgElementStatusReport):
        if not scanned_element.responsible:
            return assignees, statuses

        assignees = set(
            username for username in iter_gh_usernames_from_responsibles_mapping(
                responsibles_mapping=scanned_element.responsible,
                gh_api=gh_api,
            )
            if github.user.is_user_active(
                username=username,
                github=gh_api,
            )
        )

    else:
        raise TypeError(scanned_element)

    return assignees, statuses


def _scanned_element_title(
    scanned_element: gcm.Target,
    issue_type: str,
) -> str:
    if gcm.is_ocm_artefact_node(scanned_element):
        c = scanned_element.component
        a = gcm.artifact_from_node(scanned_element)

        return f'[{issue_type}] - {c.name}:{a.name}'

    elif isinstance(scanned_element, cmm.CfgElementStatusReport):
        return f'[{issue_type}] - {scanned_element.name}'

    else:
        raise TypeError(scanned_element)


def _scanned_element_ctx_label(
    scanned_element: gcm.Target,
) -> tuple[str]:
    if gcm.is_ocm_artefact_node(scanned_element):
        return ()

    elif isinstance(scanned_element, cmm.CfgElementStatusReport):
        digest_label = github.compliance.issue.digest_label(
            prefix=_ctx_label_prefix,
            digest_str=scanned_element.element_storage,
        )
        return (digest_label, )

    else:
        raise TypeError(scanned_element)


@functools.cache
def _valid_issue_assignees(
    repository: github3.repos.Repository,
) -> set[str]:
    return set(
        u.login for u in repository.assignees()
    )


class PROCESSING_ACTION(enum.Enum):
    DISCARD = 'discard'
    REPORT = 'report'


def create_or_update_github_issues(
    result_group_collection: gcm.ScanResultGroupCollection,
    max_processing_days: gcm.MaxProcessingTimesDays=None,
    gh_api: github3.GitHub=None,
    overwrite_repository: github3.repos.Repository=None,
    preserve_labels_regexes: typing.Iterable[str]=(),
    github_issue_template_cfgs: list[image_scan.GithubIssueTemplateCfg]=None,
    delivery_svc_client: delivery.client.DeliveryServiceClient=None,
    delivery_svc_endpoints: model.delivery.DeliveryEndpointsCfg=None,
    gh_quota_minimum: int = 2000, # skip issue updates if remaining quota falls below this threshold
):
    result_groups = result_group_collection.result_groups
    result_groups_with_findings = result_group_collection.result_groups_with_findings
    result_groups_without_findings = result_group_collection.result_groups_without_findings
    result_groups_with_scan_error = result_group_collection.result_groups_with_scan_errors

    err_count = 0

    def is_remaining_quota_too_low() -> bool:
        ratelimit_remaining = gh_api.ratelimit_remaining
        logger.info(f'{ratelimit_remaining=}')
        if ratelimit_remaining < gh_quota_minimum:
            return True
        return False

    def process_result(
        result_group: gcm.ScanResultGroup,
        action: PROCESSING_ACTION,
    ):
        nonlocal gh_api
        nonlocal err_count
        nonlocal max_processing_days
        issue_type = result_group.issue_type

        if action == PROCESSING_ACTION.DISCARD:
            results = result_group.results_without_findings
        elif action == PROCESSING_ACTION.REPORT:
            results = result_group.results_with_findings

        criticality_classification = result_group.worst_severity

        scan_result = result_group.results[0]
        ctx_labels = _scanned_element_ctx_label(scan_result.scanned_element)

        if gcm.is_ocm_artefact_node(scan_result.scanned_element):
            if results and not len({r.scanned_element.component.name for r in results}) == 1:
                raise ValueError('not all component names are identical')

        if overwrite_repository:
            repository = overwrite_repository
        else:
            repository = _scanned_element_repository(scan_result.scanned_element)

        known_issues = _all_issues(repository)

        if action == PROCESSING_ACTION.DISCARD:
            github.compliance.issue.close_issue_if_present(
                scanned_element=scan_result.scanned_element,
                issue_type=issue_type,
                repository=repository,
                known_issues=known_issues,
                ctx_labels=ctx_labels,
            )

            element_name = github.compliance.issue.unique_name_for_element(
                scanned_element=scan_result.scanned_element,
            )
            logger.info(f'closed (if existing) gh-issue for {element_name=}')

        elif action == PROCESSING_ACTION.REPORT:
            assignees, assignees_statuses = _scanned_element_assignees(
                scanned_element=scan_result.scanned_element,
                delivery_svc_client=delivery_svc_client,
                repository=repository,
                gh_api=gh_api,
            )

            valid_assignees = _valid_issue_assignees(repository)

            # Make sure all names are lowercase (as recommended by gh)
            assignees = set(a.lower() for a in assignees)
            valid_assignees = set(a.lower() for a in valid_assignees)

            if invalid_assignees := (assignees - valid_assignees):
                logger.warning(
                    f'Unable to assign {invalid_assignees} to issues in repository '
                    f'{repository.url}. Please make sure the users have the necessary permissions '
                    'to see issues in the repository.'
                )
                assignees -= invalid_assignees
                logger.info(
                    f'Removed invalid assignees {invalid_assignees} from target assignees for '
                    f'issue. Remaining assignees: {assignees}'
                )

            latest_processing_date = None
            target_milestone = None
            failed_milestones = []

            if delivery_svc_client:
                try:
                    if not scan_result.severity:
                        if not max_processing_days:
                            max_processing_days = gcm.MaxProcessingTimesDays()
                        max_days = max_processing_days.for_severity(
                            criticality_classification,
                        )
                        latest_processing_date = datetime.date.today() + datetime.timedelta(
                            days=max_days,
                        )
                    else:
                        # do not pass delivery service client or repository here to avoid
                        # determining milestone here because then we would lose track of
                        # failed milestone assignments
                        latest_processing_date = scan_result.calculate_latest_processing_date(
                            max_processing_days=max_processing_days,
                        )

                    target_sprints = gcmi.target_sprints(
                        delivery_svc_client=delivery_svc_client,
                        latest_processing_date=latest_processing_date,
                        sprints_count=2,
                    )
                    target_milestone, failed_milestones = gcmi.find_or_create_sprint_milestone(
                        repo=repository,
                        sprints=target_sprints,
                    )

                    if target_milestone:
                        latest_processing_date = target_milestone.due_on.date()
                except Exception as e:
                    import traceback
                    traceback.print_exc()
                    logger.warning(f'{e=}')
                    target_milestone = None

            if gcm.is_ocm_artefact_node(scan_result.scanned_element) and delivery_svc_endpoints:
                delivery_dashboard_url = _delivery_dashboard_url(
                    component=scan_result.scanned_element.component,
                    base_url=delivery_svc_endpoints.dashboard_url(),
                )
                delivery_dashboard_url = f'[Delivery-Dashboard]({delivery_dashboard_url})'
            else:
                delivery_dashboard_url = ''

            template_variables = _template_vars(
                result_group=result_group,
                delivery_dashboard_url=delivery_dashboard_url,
            )

            for issue_cfg in github_issue_template_cfgs:
                if issue_cfg.type == issue_type:
                    break
            else:
                raise ValueError(f'no template for {issue_type=}')

            body = issue_cfg.body.format(**template_variables)

            try:
                issue = github.compliance.issue.create_or_update_issue(
                    scanned_element=scan_result.scanned_element,
                    issue_type=issue_type,
                    repository=repository,
                    body=body,
                    assignees=assignees,
                    assignees_statuses=assignees_statuses,
                    milestone=target_milestone,
                    failed_milestones=failed_milestones,
                    latest_processing_date=latest_processing_date,
                    ctx_labels=ctx_labels,
                    preserve_labels_regexes=preserve_labels_regexes,
                    known_issues=known_issues,
                    title=_scanned_element_title(
                        scanned_element=scan_result.scanned_element,
                        issue_type=issue_type,
                    ),
                )

                element_name = github.compliance.issue.unique_name_for_element(
                    scanned_element=scan_result.scanned_element,
                )
                logger.info(
                    f'updated gh-issue for {element_name=} {issue_type=}: {issue.html_url=}'
                )
            except github3.exceptions.GitHubError as ghe:
                err_count += 1
                logger.warning('error whilst trying to create or update issue - will keep going')
                logger.warning(f'error: {ghe} {ghe.code=} {ghe.message=}')

        else:
            raise NotImplementedError(action)

    if is_remaining_quota_too_low():
        logger.warning(f'skipping issue updates, quota too low; {gh_quota_minimum=}')
        return

    for result_group in result_groups_with_findings:
        process_result(
            result_group=result_group,
            action=PROCESSING_ACTION.REPORT,
        )
        time.sleep(1) # throttle github-api-requests

    if is_remaining_quota_too_low():
        logger.warning(f'skipping issue updates, quota too low; {gh_quota_minimum=}')
        return

    for result_group in result_groups_without_findings:
        if not result_group.has_attempted_scans:
            continue
        logger.info(f'discarding issue for {result_group.name=}')
        process_result(
            result_group=result_group,
            action=PROCESSING_ACTION.DISCARD,
        )
        time.sleep(1) # throttle github-api-requests

    if result_groups_with_scan_error:
        logger.warning(f'{len(result_groups_with_scan_error)=} had scanning errors (check log)')
        # do not fail job (for now); might choose to, later

    if is_remaining_quota_too_low():
        logger.warning(f'skipping issue updates, quota too low; {gh_quota_minimum=}')
        return

    if overwrite_repository:
        known_issues = _all_issues(overwrite_repository)
        issue_type = result_group_collection.issue_type

        all_ctx_labels = set()

        if result_groups:
            for result_group in result_groups:
                scanned_element = result_group.results[0].scanned_element
                if (ctx_labels := _scanned_element_ctx_label(scanned_element)):
                    all_ctx_labels = all_ctx_labels | set(ctx_labels)

        else:
            logger.info('no scan results, will skip issues with ctx label')

        close_issues_for_absent_or_assessed_resources(
            result_groups=result_groups,
            known_issues=known_issues,
            issue_type=issue_type,
            ctx_labels=all_ctx_labels,
        )

    if err_count > 0:
        logger.warning(f'{err_count=} - there were errors - will raise')
        raise RuntimeError('not all github-issues could be created/updated/deleted')

    logger.info(f'{gh_api.ratelimit_remaining=}')


def close_issues_for_absent_or_assessed_resources(
    result_groups: list[gcm.ScanResultGroup],
    known_issues: typing.Iterator[github3.issues.issue.ShortIssue],
    issue_type: str | None,
    ctx_labels: typing.Iterable[str]=(),
):
    '''
    closes all open issues for scanned elements that are not present in given result-groups.

    this is intended to automatically close issues for scan targets that are no longer present.
    '''

    def close_issues(
        issues: typing.Iterable[github3.issues.Issue],
        resources_in_bom: set[str]=set(),
    ):
        for issue in issues:
            logger.info(
                f"Closing issue '{issue.title}'({issue.html_url}) since no scan contained a "
                "scanned element matching its digest."
            )
            comment = (
                'closing ticket, because scanned element is no longer present in BoM ' +
                'or there are no longer unassessed findings'
            )
            for resource_in_bom in resources_in_bom:
                if resource_in_bom in issue.title:
                    # if there is still an issue open for this resource, it is still present in BoM
                    comment = 'closing ticket, because there are no longer unassessed findings'
                    break
            issue.create_comment(comment)
            github.util.close_issue(issue)

    def has_ctx_label(
        issue: github3.issues.Issue,
    ) -> bool:
        return any([
            l.name.startswith(_ctx_label_prefix)
            for l in issue.original_labels]
        )

    all_issues = github.compliance.issue.enumerate_issues(
        scanned_element=None,
        issue_type=issue_type,
        known_issues=known_issues,
        state='open',
        extra_labels=ctx_labels,
    )

    if not ctx_labels:
        all_issues = (
            issue
            for issue in all_issues
            if not has_ctx_label(issue)
        )

    if not result_groups:
        logger.info(f'no scan results, will close all issues for {issue_type=} and {ctx_labels=}')
        close_issues(all_issues)
        return

    scanned_element = result_groups[0].results[0].scanned_element
    prefix = github.compliance.issue.prefix_for_element(scanned_element)

    def component_resource_label(issue: github3.issues.Issue) -> str:
        for label in issue.original_labels:
            label: github3.issues.label.ShortLabel
            if label.name.startswith(prefix):
                return label.name

    component_resources_to_issues = {
        component_resource_label(issue): issue for issue in all_issues
    }

    resources_in_bom = set()
    for result_group in result_groups:
        scanned_element = result_group.results[0].scanned_element

        name = github.compliance.issue.unique_name_for_element(
            scanned_element=scanned_element,
        )
        prefix = github.compliance.issue.prefix_for_element(scanned_element)

        resource_label = github.compliance.issue.digest_label(
            prefix=prefix,
            digest_str=name,
        )

        logger.info(f'Digest-Label for {result_group.name=}: {resource_label=}')
        component_resources_to_issues.pop(resource_label, None)

        resources_in_bom.add(name)

    # any issues that have not been removed thus far were not referenced by given result_groups
    close_issues(
        issues=component_resources_to_issues.values(),
        resources_in_bom=resources_in_bom,
    )



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0

from model.base import NamedModelElement

from . import cluster_domain_from_kubernetes_config

OAUTH_PROXY_SUBDOMAIN_LABEL = 'tkn-auth'


class Oauth2ProxyConfig(NamedModelElement):
    '''Not intended to be instantiated by users of this module
    '''

    def oauth2_proxy_chart_config(self):
        return Oauth2ProxyChartConfig(
            '',
            self.raw.get('oauth2_proxy_chart_config')
        )

    def github_oauth_config(self):
        return GithubOauthConfig(
            '',
            self.raw.get('github_oauth_config')
        )

    def external_url(self, cfg_factory):
        return self.ingress_host(cfg_factory)

    def ingress_config(self):
        return self.raw.get('ingress_config')

    def ingress_host(self, cfg_factory):
        cluster_domain = cluster_domain_from_kubernetes_config(
            cfg_factory,
            self.kubernetes_config_name(),
        )
        return f'{self.subdomain_label()}.{cluster_domain}'

    def namespace(self):
        return self.raw.get('namespace')

    def kubernetes_config_name(self):
        return self.raw.get('kubernetes_config')

    def subdomain_label(self):
        return self.raw.get('subdomain_label', OAUTH_PROXY_SUBDOMAIN_LABEL)

    def _required_attributes(self):
        yield from super()._required_attributes()
        yield from [
            'github_oauth_config',
            'ingress_config',
            'kubernetes_config',
            'namespace',
            'oauth2_proxy_chart_config',
        ]

    def _optional_attributes(self):
        yield from super()._optional_attributes()
        yield from [
            'subdomain_label',
        ]


class Oauth2ProxyChartConfig(NamedModelElement):
    def cookie_secret(self):
        return self.raw.get('cookie_secret')

    def cookie_name(self):
        return self.raw.get('cookie_name')

    def whitelist_domains(self):
        return self.raw.get('whitelist_domains')

    def cookie_domain(self):
        return self.raw.get('cookie_domain')

    def _required_attributes(self):
        yield from super()._required_attributes()
        yield from [
            'cookie_secret',
        ]

    def _optional_attributes(self):
        yield from super()._optional_attributes()
        yield from [
            'cookie_domain',
            'cookie_name',
            'whitelist_domains',
        ]


class GithubOauthConfig(NamedModelElement):
    # Move to github.py?

    def client_id(self):
        return self.raw.get('client_id')

    def client_secret(self):
        return self.raw.get('client_secret')

    def github_cfg_name(self):
        return self.raw.get('github_cfg_name')

    def github_org(self):
        return self.raw.get('github_org')

    def github_team(self):
        return self.raw.get('github_team')

    def no_ssl_verify(self):
        return self.raw.get('no_ssl_verify', False)

    def _required_attributes(self):
        yield from super()._required_attributes()
        yield from [
            'client_id',
            'client_secret',
            'github_cfg_name',
            'github_org',
            'github_team',
        ]

    def _optional_attributes(self):
        yield from super()._optional_attributes()
        yield from [
            'no_ssl_verify',
        ]



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import dataclasses

import ci.util


class ModelValidationError(ValueError):
    '''
    exception to be raised upon model validation errors
    '''
    pass


class ConfigElementNotFoundError(ValueError):
    pass


class ModelValidationMixin:
    def _required_attributes(self):
        return ()

    def _optional_attributes(self):
        return ()

    def _known_attributes(self):
        return set(self._required_attributes()) | \
                set(self._optional_attributes()) | \
                set(self._defaults_dict().keys())

    def validate(self):
        self._validate_required_attributes()
        self._validate_known_attributes()

    def _validate_required_attributes(self):
        missing_attributes = [a for a in self._required_attributes() if a not in self.raw]
        if missing_attributes:
            if hasattr(self, 'name'):
                if callable(self.name):
                    name = self.name()
                else:
                    name = str(self.name)
            else:
                name = '<unknown>'

            raise ModelValidationError(
                f'{type(self).__name__}:{name}: The following required attributes are '
                f'absent: {", ".join(missing_attributes)}.'
            )

    def _validate_known_attributes(self):
        unknown_attributes = [a for a in self.raw if a not in self._known_attributes()]
        if unknown_attributes:
            if hasattr(self, 'name'):
                if callable(self.name):
                    name = self.name()
                else:
                    name = str(self.name)
            else:
                name = '<unknown>'

            raise ModelValidationError(
                f'{type(self).__name__}:{name}: The following attributes are '
                f'unknown: {", ".join(unknown_attributes)}.'
            )


class ModelDefaultsMixin:
    def _defaults_dict(self):
        return {}

    def _apply_defaults(self, raw_dict):
        self.raw = ci.util.merge_dicts(
            self._defaults_dict(),
            raw_dict,
        )


class ModelBase(ModelValidationMixin, ModelDefaultsMixin):
    '''
    Base class for 'dict-based' configuration classes (i.e. classes that expose contents
    from a dict through a set of 'getter' methods.

    Extenders _may_ overwrite `_required_attributes(self)` and return an iterable of attribute
    identifiers. If such an iterable is returned, the ctor ensures that all specified attributes be
    contained in the given dictionary (ModelValidationError is raised on absent attribs).
    '''

    def __init__(self, raw_dict):
        self.raw = ci.util.not_none(raw_dict)

    def __repr__(self):
        return '{c} {a}'.format(
            c=self.__class__.__name__,
            a=str(self.raw),
        )


class NamedModelElement(ModelBase):
    def __init__(
        self,
        name: str,
        raw_dict: dict,
        type_name: str=None,
        *args,
        **kwargs
    ):
        self._name = ci.util.not_none(name)
        self._type_name = type_name
        super().__init__(raw_dict=raw_dict, *args, **kwargs)

    def _optional_attributes(self):
        # workaround: NamedModelElement allows any attribute; it would
        # obviously be a better way to disable this check (e.g. split into
        # separate mixin and not add it to NME
        return set(self.raw.keys())

    def name(self):
        return self._name

    def __getattr__(self, name):
        if name in self.raw:
            return lambda: self.raw[name]

        raise AttributeError(name)

    def __repr__(self):
        return f'{self.__class__.__qualname__}: {self.name()}'

    def __str__(self):
        return '{n}: {d}'.format(n=self.name(), d=self.raw)


class BasicCredentials(ModelBase):
    '''
    Base class for configuration objects that contain basic authentication credentials
    (i.e. a username and a password)

    Not intended to be instantiated
    '''

    def username(self):
        return self.raw.get('username')

    def passwd(self):
        return self.raw.get('password')

    def as_tuple(self):
        return (self.username(), self.passwd())

    def _required_attributes(self):
        return ['username', 'password']


class TokenCredentials(ModelBase):
    '''
    Base class for configuration objects that use token-based authentication

    Not intended to be instantiated
    '''

    def token(self):
        return self.raw['token']

    def _required_attributes(self):
        return ['token']


@dataclasses.dataclass(frozen=True)
class CfgElementReference:
    type_name: str
    element_name: str
    purpose: str | None = None



import collections.abc
import dataclasses
import datetime
import functools
import logging

import github3.repos

import delivery.client
import delivery.model as dm


logger = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class MilestoneConfiguration:
    title_callback: collections.abc.Callable[[dm.Sprint], str] = lambda sprint: sprint.name
    title_prefix: str | None = 'sprint-'
    title_suffix: str | None = None
    due_date_callback: collections.abc.Callable[[dm.Sprint], datetime.datetime] \
        = lambda sprint: sprint.find_sprint_date('release_decision').value


def _milestone_title(
    sprint: dm.Sprint,
    milestone_cfg: MilestoneConfiguration=None,
) -> str:
    if not milestone_cfg:
        milestone_cfg = MilestoneConfiguration()

    title = milestone_cfg.title_callback(sprint)
    title_prefix = milestone_cfg.title_prefix or ''
    title_suffix = milestone_cfg.title_suffix or ''

    return f'{title_prefix}{title}{title_suffix}'


def find_sprint_milestone(
    repo: github3.repos.Repository,
    sprints: tuple[dm.Sprint],
    milestone_cfg: MilestoneConfiguration=None,
) -> tuple[
    dm.Sprint,
    github3.repos.repo.milestone.Milestone,
    list[github3.repos.repo.milestone.Milestone],
]:
    sprint_milestone = None
    failed_milestones = []
    all_milestones = repo.milestones(state='all')

    for sprint in sprints:
        for ms in all_milestones:
            if ms.title == _milestone_title(
                sprint=sprint,
                milestone_cfg=milestone_cfg,
            ):
                sprint_milestone = ms
                break

        if not sprint_milestone:
            # milestone does not exist yet -> create it
            return (sprint, None, failed_milestones)

        if sprint_milestone.state == 'open':
            # milestone exists and is open -> use it
            return (sprint, sprint_milestone, failed_milestones)

        # milestone exists but is closed -> repeat with next sprint
        failed_milestones.append(sprint_milestone)
        sprint_milestone = None

    return (None, None, failed_milestones)


@functools.cache
def find_or_create_sprint_milestone(
    repo: github3.repos.Repository,
    sprints: tuple[dm.Sprint],
    milestone_cfg: MilestoneConfiguration=None,
) -> tuple[
    github3.repos.repo.milestone.Milestone | None,
    list[github3.repos.repo.milestone.Milestone],
]:
    if not milestone_cfg:
        milestone_cfg = MilestoneConfiguration()

    sprint, milestone, failed_milestones = find_sprint_milestone(
        repo=repo,
        sprints=sprints,
        milestone_cfg=milestone_cfg,
    )

    if milestone:
        return (milestone, failed_milestones)

    if not sprint:
        # all sprints have sprint milestones which were extraordinary closed
        # because sprints are still in the future but there milestone is closed
        # we can't do anything here so write info to ticket
        return (None, failed_milestones)

    title = _milestone_title(
        sprint=sprint,
        milestone_cfg=milestone_cfg,
    )

    due_date = milestone_cfg.due_date_callback(sprint)

    ms = repo.create_milestone(
        title=title,
        state='open',
        description=f'used to track issues for upcoming sprint {title}',
        due_on=due_date.isoformat(),
    )

    return (ms, failed_milestones)


@functools.cache
def _sprints(
    delivery_svc_client: delivery.client.DeliveryServiceClient,
    today: datetime.date=datetime.date.today(), # used to refresh the cache daily
) -> list[dm.Sprint]:
    return delivery_svc_client.sprints()


def target_sprints(
    delivery_svc_client: delivery.client.DeliveryServiceClient,
    latest_processing_date: datetime.date,
    sprints_count: int=1,
) -> tuple[dm.Sprint]:
    sprints = _sprints(
        delivery_svc_client=delivery_svc_client,
    )
    sprints.sort(key=lambda sprint: sprint.find_sprint_date(name='end_date').value.date())

    targets_sprints = []
    for sprint in sprints:
        if len(targets_sprints) == sprints_count:
            # found enough sprints -> early exiting
            break

        end_date = sprint.find_sprint_date(name='end_date').value.date()
        if end_date >= latest_processing_date:
            targets_sprints.append(sprint)

    if len(targets_sprints) < sprints_count:
        logger.warning(
            f'did not find {sprints_count} sprints starting from ' +
            f'{latest_processing_date}, only found {len(targets_sprints)} sprints'
        )

    return tuple(targets_sprints)



# SPDX-FileCopyrightText: 2021 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0

import sys
import os

own_dir = os.path.abspath(os.path.dirname(__name__))
repo_root = os.path.abspath(os.path.join(own_dir, os.path.pardir))

sys.path.insert(1, repo_root)
